/*---------------------------------------------------------------------------------------------------------------------*
 * UNIVERSIDADE DO VALE DO RIO DOS SINOS - MESTRADO EM COMPUTACAO APLICADA                                             *
 * Sistemas Adaptativos Inteligentes - Prof. Fernando Santos Osorio                                                    *
 * Autor: Milton Roberto Heinen - Matr. 392201                                                                         *
 *---------------------------------------------------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------------------------------------------------*
 * Includes                                                                                                            *
 *---------------------------------------------------------------------------------------------------------------------*/
#include <ode/ode.h>
#include <drawstuff/drawstuff.h>
#include <ga/ga.h>
#define INSTANTIATE_REAL_GENOME
#include <ga/GARealGenome.h>
#include "recipes.h"
#include <signal.h>
#include <string>
using std::string;


/*---------------------------------------------------------------------------------------------------------------------*
 * Constantes                                                                                                          *
 *---------------------------------------------------------------------------------------------------------------------*/
// Constantes fisicas
const int MAX_CONTACTS = 4;
const int NUM_DIMENSOES = 3;
const double GRAVITY = -9.81;
const double ERP = 0.8;
const double CFM = 1.0e-4;
const double DENSIDADE = 0.5;
const double FORCA_MAX = 3500.0;
const double PASSO_FISICO = 0.05;

// Constantes da janela de visualizacao
const int TAM_TELA_X = 640;
const int TAM_TELA_Y = 480;

// Constantes realativas as velocidades e ao automato
const double TEMPO_MAX_ESTADO = 1.0;
const double INTERVALO_POSICAO = 0.01;
const double INTERVALO_VELOCIDADE = 0.01;
const double VELOC_RELATIVA_MIN = 0.25;
const double VELOC_RELATIVA_MAX = 1.75;

// Constantes dos sensores do robo
const double SAMPLE_GYRO_INTERVAL = 0.5;
const int MAX_SAMPLES_GYRO = 1024;
const int MAX_SAMPLES_BUMPERS = 10240;
const double PESO_GYRO = 10.0;

// Constantes dos alelos do rob
const double TAM_MAX_CORPO = 5.0;
const double TAM_MAX_PARTES = 2.5;
const double INTERVALO_ALELOS_MORFO = 0.01;

// Constantes gerais
const double NOVENTA_GRAUS_RAD = M_PI / 2.0;
const double TOLERANCIA_POWELL = 1.0e-5;
const int MAX_LINHA = 5120;


/*---------------------------------------------------------------------------------------------------------------------*
 * Variaveis globais                                                                                                   *
 *---------------------------------------------------------------------------------------------------------------------*/
// Mundo fisico
dWorldID World = NULL;
dBodyID *bodyMembros = NULL;
dBodyID *bodyJuntas = NULL;
dBodyID bodyCorpo;

// Espaca colisivo
dSpaceID Space = NULL;
dJointGroupID ContactGroup;
dGeomID *geomMembros = NULL;
dGeomID geomCorpo;
dGeomID geomSolo;

// Juntas
dJointID *hingeMembros = NULL;
dJointID *motorMembros = NULL;
dJointID *fixedMembros = NULL;

// Variaveis de drawstuff
dsFunctions myDrawFunction;
float CAMERA_XYZ[2][NUM_DIMENSOES] = {{  4.0f, 8.5f, 3.0f}, {  13.0f, 0.0f, 2.0f}};
float CAMERA_HPR[2][NUM_DIMENSOES] = {{-90.0f, 0.0f, 0.0f}, {-180.0f, 0.0f, 0.0f}};
char PATH_TEXTURES[] = "./textures";

// Variaveis do robotnik
string sNomeArquivoBase = "";
int iNumPernasRobotnik = 0;
int iNumPartesPorPerna = 0;
int iTotalPartesPernas = 0;
double **ppdTamPartesPernas = NULL;
double **ppdPosPartesPernas = NULL;
double **ppdPosJuntasPernas = NULL;
double **ppdRestricoesJunta = NULL;
double vdTamCorpoRobot[NUM_DIMENSOES];
double vdPosCorpoRobot[NUM_DIMENSOES];
bool bAngulosPatasCalculados = false;

// Variaveis sensoriais
double vdGyroSamplesX[MAX_SAMPLES_GYRO];
double vdGyroSamplesY[MAX_SAMPLES_GYRO];
double vdGyroSamplesZ[MAX_SAMPLES_GYRO];
bool **pbBumperSamples = NULL;
bool *pbContatoComSolo = NULL;
int iCountGiroSamples = 0;
int iCountBumperSamples = 0;

// Variaveis da simulacao
bool bEcoTelaHabilitado = false;
bool bAplicacaoCancelada = false;
bool bSimulacaoVisual = true;
double dTimeSimulation = 0.0;
long lRandomSeed = 0;

// Variaveis do Automato
int iNumEstados = 0;
double *pdVelocRelativa = NULL;
double **ppdAngleAutomato = NULL;
double **ppdVelocAutomato = NULL;
int iEstadoAtual = -1;

// Variaveis geneticas
long lGeracaoAtual = 0;
long lEpochOfBest = 0;

// Variaveis de Powell
int iPernaAtualCinemat = 0;
double dPosDesejadaX = 0.0;
double dPosDesejadaY = 0.0;
double *pdAnglesCinemat = NULL;
double **ppdMatrixIdenty = NULL;


/*---------------------------------------------------------------------------------------------------------------------*
 * Prototipos                                                                                                          *
 *---------------------------------------------------------------------------------------------------------------------*/
// Funcoes estaticas ODE
static void NearCallback(void *xxx, dGeomID o1, dGeomID o2);
static void DrawstuffStart();
static void DrawstuffCommand(int cmd);
static void DrawstuffSimLoop(int pause);

// Funcoes do Robotnik
void AlocarMemoriaRobotnik(const int iNumPernas, const int iNumPartes);
void CarregarRobotnikFromFile(const string sNomeArquivo);
void CalcularPosicoesRobotnik();
void SalvarRobotnik(const string sNomeArquivo);
void AlocarMemoriaAutomato();
void LiberarMemoriaRobotnik();
void LiberarMemoriaAutomato();

// Mundo fisico e colisivo
void CriaUniverso();
void CriaTodosObjetos();
void CriaJuntas();
void DestruirUniverso();

// Funcoes de controle da simulacao
void SetAutomatoFromFile(const string sNomeArquivo);
void SetaPosOriginal();
double PhysicalSimLoop(const double dTimeSimulation);
void PararTodasAsJuntas();
inline void AjustarVelocidades();
void CalcularVelocidades();
void AjustarAngulosPatas();
void VerificarRestricoes();
void GravarAutomato(const string sNomeArquivo);

// Funcoes da cinematica inversa
void AlocarMemoriaPowell();
inline void InicializarMatrizIdentidade();
static double CinematicaInversa(double dAngles[]);
inline double SimplificarAngulo(const double dAngle);
double CalcularTamanhoPerna(const int iPerna);
void AjustaDistanciaPonto(double &dPosX, double &dPosY);
bool SetTudinhoFromGenoma(GAGenome & g);
void LiberarMemoriaPowell();

// Funcoes sensoriais
double GetAlturaMaxBumpers();
double GetAlturaMinBumpers();
inline bool AlturaPataOk(const int iNumeroPata);
double CalcularInstabilidade();
double CalcularFatorBumpers();
double CalcularDeslocamento();

// Funcoes do algoritmo genetico
void CarregarParametros(const string sArqParams);
void AlgoritmoGenetico(const string sArqParams);
static void Initializer(GAGenome& g);
static float Objective(GAGenome & g);
bool SetRobotnikFromGenoma(GAGenome & g);

// Funcoes Utilitarias
string GetParamInFile(const string sNameFile, const string sParam);
inline double DegToRad(const double dValDeg);
inline double RadToDeg(const double dValRad);
inline double Quadrado(const double dValor);
inline double Sinal(const double dValor);
inline double Menor(const double dValorA, const double dValorB);
void HabilitaEcoTela(int iSinal);

// Funcoes de finalizacao
void EncerrarSimulacao();
void TrataPedidoDestruicao(int iSinal);


/*---------------------------------------------------------------------------------------------------------------------*
 * Funcao main                                                                                                         *
 *---------------------------------------------------------------------------------------------------------------------*/
int main(int argc, char *argv[])
{
  // Configura a manipulacao dos sinais de cancelamento
  struct sigaction saCancel;
  memset(&saCancel, 0, sizeof(saCancel));
  saCancel.sa_handler = &TrataPedidoDestruicao;
  sigaction(SIGINT, &saCancel, NULL);  
  sigaction(SIGTERM, &saCancel, NULL);
  
  // Configura a manipulacao dos sinais do usuario (eco)
  struct sigaction saSigusr;
  memset(&saSigusr, 0, sizeof(saSigusr));
  saSigusr.sa_handler = &HabilitaEcoTela;
  sigaction(SIGUSR1, &saSigusr, NULL);
  sigaction(SIGUSR2, &saSigusr, NULL);
  
  // Le os parametros da linha de comando
  if (argc > 1 && argv[1] && strncmp(argv[1], "-", 1))
    sNomeArquivoBase = argv[1];
  else {
    printf("Uso: %s <nome_arquivo> [-l]\n", argv[0]);
    return 1;
  }
  if (argc > 2 && argv[2] && !strcmp(argv[2], "-l"))
    bSimulacaoVisual = false;
    
  // Cria o grupo de contatos
  ContactGroup = dJointGroupCreate(0);

  if (bSimulacaoVisual) {
    // Visulizacao da simulacao
    CarregarRobotnikFromFile(sNomeArquivoBase + ".rbt");
    CriaUniverso();
    CriaTodosObjetos();
    CriaJuntas();
    SetAutomatoFromFile(sNomeArquivoBase + ".ctr");
    dsSimulationLoop(argc, argv, TAM_TELA_X, TAM_TELA_Y, &myDrawFunction);
  }
  else {
    // Aprendizado genetico
    CarregarRobotnikFromFile(sNomeArquivoBase + "_original.rbt");
    CarregarParametros(sNomeArquivoBase + ".par");
    AlocarMemoriaAutomato();
    AlocarMemoriaPowell();
    AlgoritmoGenetico(sNomeArquivoBase + ".par");
  }
  EncerrarSimulacao();
  return 0;
}


/*---------------------------------------------------------------------------------------------------------------------*
 * Funcoes estaticas ODE                                                                                               *
 *---------------------------------------------------------------------------------------------------------------------*/
static void NearCallback(void *xxx, dGeomID o1, dGeomID o2)
{
  static dContact contact[MAX_CONTACTS];
  register int i, iNumContacts = 0;
  static dBodyID b1, b2;
  static dJointID c;

  // Verifica se estao conectados
  b1 = dGeomGetBody(o1);
  b2 = dGeomGetBody(o2);
  if (b1 && b2 && dAreConnected(b1, b2))
    return;

  // Busca o numero de pontos de contato
  if (!(iNumContacts = dCollide(o1, o2, MAX_CONTACTS, &contact[0].geom, sizeof(dContact))))
    return;

  // Verifica se as patas entraram em contato com o solo neste turno (bumpers)
  if (o1 == geomSolo || o2 == geomSolo) {
    for (i = 0; i < iNumPernasRobotnik; i++) {
      if (!pbContatoComSolo[i] && (o1 == geomMembros[i * iNumPartesPorPerna] || o2  == geomMembros[i * iNumPartesPorPerna])) {
        pbContatoComSolo[i] = true;
        break;
      }
    }
  }

  for (i = 0; i < iNumContacts; i++) {
    // Configura a superficie de contato
    contact[i].surface.mode = dContactBounce | dContactSoftCFM;
    contact[i].surface.mu = dInfinity;
    contact[i].surface.mu2 = 0.0;
    contact[i].surface.bounce = 0.001;
    contact[i].surface.bounce_vel = 0.1;
    contact[i].surface.soft_cfm = 0.0001;

    // Cria as juntas do tipo contato para os pontos de colisao
    c = dJointCreateContact(World, ContactGroup, contact + i);
    dJointAttach(c, b1, b2);
  }
}


static void DrawstuffStart()
{
  // Inicializa a janela visual - seta a posicao da camera
  dsSetViewpoint(CAMERA_XYZ[0], CAMERA_HPR[0]);
}


static void DrawstuffCommand(int cmd)
{
  // Interpretador de comandos do teclado
  if (cmd == 'x' || cmd == 'X') {
    PararTodasAsJuntas();
    EncerrarSimulacao();
  }
  else if (cmd == 'r' || cmd == 'R') {
    PararTodasAsJuntas();
    SetaPosOriginal();
  }
  else if (cmd == 's' || cmd == 'S') {
    if (iEstadoAtual < 0)
      iEstadoAtual = 0;
  }
  else if (cmd == 'v' || cmd == 'V')
    dsSetViewpoint(CAMERA_XYZ[0], CAMERA_HPR[0]);
  else if (cmd == 'f' || cmd == 'F')
    dsSetViewpoint(CAMERA_XYZ[1], CAMERA_HPR[1]);
}


static void DrawstuffSimLoop(int pause)
{
  register int i;
  struct timespec ts;
  static double dRelogioSimulacao = 0.0;
  static clock_t clAnterior = clock();
  clock_t clAtual = clock();
  dVector3 sizeBox;

  // Da uma pausa para que o tempo de exibicao se equipare ao tempo real de simulacao
  ts.tv_sec = 0;
  ts.tv_nsec = (long) ((PASSO_FISICO - (double) (clAtual - clAnterior) / (double) CLOCKS_PER_SEC) * 1000000000.0);
  nanosleep(&ts, NULL);
  clAnterior = clock();

  // Laco da simulacao fisico
  if (!pause && iEstadoAtual >= 0) {
    AjustarVelocidades();
    dSpaceCollide(Space, 0, NearCallback);
    dWorldStep(World, PASSO_FISICO);
    dJointGroupEmpty(ContactGroup);
    dRelogioSimulacao += PASSO_FISICO;
  }

  // Desenha o robo
  dsSetTexture(DS_WOOD);
  dsSetColor(1, 1, 0);
  // Mebros
  for (i = 0; i < iTotalPartesPernas; i++) {
    dGeomBoxGetLengths(geomMembros[i], sizeBox);
    dsDrawBoxD(dBodyGetPosition(bodyMembros[i]), dBodyGetRotation(bodyMembros[i]), sizeBox);
    dsDrawCylinderD(dBodyGetPosition(bodyJuntas[i]), dBodyGetRotation(bodyJuntas[i]), Menor(ppdTamPartesPernas[i][1], (i+1) % 
        iNumPartesPorPerna ? ppdTamPartesPernas[i+1][1] : vdTamCorpoRobot[1]), Menor(Menor(ppdTamPartesPernas[i][0]/2.0, ((i+1) 
        % iNumPartesPorPerna ? ppdTamPartesPernas[i+1][0] : vdTamCorpoRobot[0])/2.0), Menor(ppdTamPartesPernas[i][2]/2.0, 
        ((i+1) % iNumPartesPorPerna ? ppdTamPartesPernas[i+1][2] : vdTamCorpoRobot[2])/2.0)));
  }
  // Corpo
  dGeomBoxGetLengths(geomCorpo, sizeBox);
  dsDrawBoxD(dBodyGetPosition(bodyCorpo), dBodyGetRotation(bodyCorpo), sizeBox);

  // Exibe o tempo de simulacao na tela
  printf("\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b");
  printf("Tempo: %02d:%02d   -   Deslocamento:%6.2f", (int) dRelogioSimulacao % 3600 / 60, (int) dRelogioSimulacao % 60,
      CalcularDeslocamento());
  fflush(stdout);
}


/*---------------------------------------------------------------------------------------------------------------------*
 * Funcoes do Robotnik                                                                                                 *
 *---------------------------------------------------------------------------------------------------------------------*/
void AlocarMemoriaRobotnik(const int iNumPernas, const int iNumPartes) 
{
  int i;
  
  // Aloca memoria para o robotnik
  pbContatoComSolo = new bool[iNumPernas];
  ppdTamPartesPernas = new double*[iNumPernas * iNumPartes];
  ppdPosPartesPernas = new double*[iNumPernas * iNumPartes];
  ppdPosJuntasPernas = new double*[iNumPernas * iNumPartes];
  ppdRestricoesJunta = new double*[iNumPernas * iNumPartes];
  for (i = 0; i < iNumPernas * iNumPartes; i++) {
    ppdTamPartesPernas[i] = new double[NUM_DIMENSOES];
    ppdPosPartesPernas[i] = new double[NUM_DIMENSOES];
    ppdPosJuntasPernas[i] = new double[NUM_DIMENSOES];
    ppdRestricoesJunta[i] = new double[2];
  }
 
  // Aloca memoria para os bumpers
  pbBumperSamples = new bool*[MAX_SAMPLES_BUMPERS];
  for (i = 0; i < MAX_SAMPLES_BUMPERS; i++)
    pbBumperSamples[i] = new bool[iNumPernas];
}
 
 
void CarregarRobotnikFromFile(const string sNomeArquivo)
{
  char vcLinha[MAX_LINHA + 1];
  char *szPalavra = NULL;
  FILE *fp = NULL;
  int i, j, k;

  // Le o arquivo com as especificacoes do robotnik
  if ((fp = fopen(sNomeArquivo.c_str(), "r")) == NULL) {
    printf("Arquivo do robotnik \'%s\' inexistente\n", sNomeArquivo.c_str());
    exit(1);
  }

  // Le a primeira linha e busca o numero de pernas, o numero de partes por perna e o sincronismo
  fgets(vcLinha, MAX_LINHA, fp);
  if ((szPalavra = strtok(vcLinha, " ")) == NULL) {
    printf("Arquivo do robotnik \'%s\' incorreto\n", sNomeArquivo.c_str());
    exit(1);
  }
  iNumPernasRobotnik = atoi(szPalavra);
  if ((szPalavra = strtok('\0', " ")) == NULL) {
    printf("Arquivo do robotnik \'%s\' incorreto\n", sNomeArquivo.c_str());
    exit(1);
  }
  iNumPartesPorPerna = atoi(szPalavra);
  iTotalPartesPernas = iNumPartesPorPerna * iNumPernasRobotnik;
  if (szPalavra && (szPalavra = strtok('\0', " ")) != NULL && strlen(szPalavra) && (szPalavra[0] == 'S' || szPalavra[0] == 's'))
    bAngulosPatasCalculados = true;

  // Aloca memoria para o robotnik
  AlocarMemoriaRobotnik(iNumPernasRobotnik, iNumPartesPorPerna);
  
  // Le a segunda linha e pega o tamanho do corpo do robotnik
  fgets(vcLinha, MAX_LINHA, fp);
  szPalavra = strtok(vcLinha, " ");
  for (i = 0; i < NUM_DIMENSOES && szPalavra != NULL; i++) {
    vdTamCorpoRobot[i] = atof(szPalavra);
    szPalavra = strtok('\0', " ");
  }

  // Le a linha subsequente para pegar o tamanho das partes das pernas do robotnik
  fgets(vcLinha, MAX_LINHA, fp);
  szPalavra = strtok(vcLinha, " ");
  for (j = 0; j < iNumPartesPorPerna && szPalavra != NULL; j++) {
    for (k = 0; k < NUM_DIMENSOES && szPalavra != NULL; k++) {
      for (i = 0; i < iNumPernasRobotnik; i++)
        ppdTamPartesPernas[i * iNumPartesPorPerna + j][k] = atof(szPalavra);
      szPalavra = strtok('\0', " ");
    }
  }

  // Le as ultimas linhas para pegar as restricoes das juntas do robotnik
  fgets(vcLinha, MAX_LINHA, fp);
  szPalavra = strtok(vcLinha, " ");
  for (j = 0; j < iNumPartesPorPerna && szPalavra != NULL; j++) {
    for (k = 0; k < 2 && szPalavra != NULL; k++) {
      for (i = 0; i < iNumPernasRobotnik; i++)
        ppdRestricoesJunta[i * iNumPartesPorPerna + j][k] = atof(szPalavra);
      szPalavra = strtok('\0', " ");
    }
  }
  fclose(fp);
  
  // Calcula as posicoes do robotnik
  CalcularPosicoesRobotnik();
}


void CalcularPosicoesRobotnik()
{
  int iCount, i, j, k; 
    
  // Calcula a posicao do corpo do robotnik
  vdPosCorpoRobot[0] = vdPosCorpoRobot[1] = 0.0;
  vdPosCorpoRobot[2] = vdTamCorpoRobot[2] / 2.0;
  for (j = 0; j < iNumPartesPorPerna; j++)
    vdPosCorpoRobot[2] += ppdTamPartesPernas[j][2];

  // Preenche as posicoes das partes do robotnik
  iCount = 0;
  for (i = 0; i < iNumPernasRobotnik; i++) {
    for (j = 0; j < iNumPartesPorPerna; j++) {
      // Coordenada x
      if ((iNumPernasRobotnik == 4 && i % 2) || (iNumPernasRobotnik == 6 && (i == 0 || i == 3)))
        ppdPosPartesPernas[iCount][0] = vdTamCorpoRobot[0] / 2.0 - ppdTamPartesPernas[i * iNumPartesPorPerna +
            iNumPartesPorPerna - 1][0] / 2.0;
      else if ((iNumPernasRobotnik == 4) || (iNumPernasRobotnik == 6 && (i == 2 || i == 5)))
        ppdPosPartesPernas[iCount][0] = -vdTamCorpoRobot[0] / 2.0 + ppdTamPartesPernas[i * iNumPartesPorPerna +
            iNumPartesPorPerna - 1][0] / 2.0;
      else if (iNumPernasRobotnik == 6 || iNumPernasRobotnik == 2)
        ppdPosPartesPernas[iCount][0] = 0;

      // Coordenada y
      if (i < iNumPernasRobotnik / 2)
        ppdPosPartesPernas[iCount][1] = vdTamCorpoRobot[1] / 2.0 - ppdTamPartesPernas[i * iNumPartesPorPerna +
            iNumPartesPorPerna - 1][1] / 2.0;
      else
        ppdPosPartesPernas[iCount][1] = -vdTamCorpoRobot[1] / 2.0 + ppdTamPartesPernas[i * iNumPartesPorPerna +
            iNumPartesPorPerna - 1][1] / 2.0;
      // Coordenada z
      ppdPosPartesPernas[iCount][2] = 0.0;
      for (k = 0; k < j; k++)
        ppdPosPartesPernas[iCount][2] += ppdTamPartesPernas[i * iNumPartesPorPerna + k][2];
      ppdPosJuntasPernas[iCount][2] = ppdPosPartesPernas[iCount][2] + ppdTamPartesPernas[i * iNumPartesPorPerna + j][2];
      ppdPosPartesPernas[iCount][2] += ppdTamPartesPernas[i * iNumPartesPorPerna + j][2] / 2.0;
      // Juntas - x e y
      for (k = 0; k < NUM_DIMENSOES - 1; k++)
        ppdPosJuntasPernas[iCount][k] = ppdPosPartesPernas[iCount][k];
      iCount++;
    }
  }
}


void SalvarRobotnik(const string sNomeArquivo)
{
  FILE *fp = NULL;
  int i, j;

  // Le o arquivo com as especificacoes do robotnik
  if ((fp = fopen(sNomeArquivo.c_str(), "w")) == NULL) {
    printf("Nao foi possivel criar o arquivo \'%s\'\n", sNomeArquivo.c_str());
    exit(1);
  }
  
  // Grava o numero de pernas, o numero de partes por perna e se os angulos das patas sao calculados
  fprintf(fp, "%d %d %s\n", iNumPernasRobotnik, iNumPartesPorPerna, bAngulosPatasCalculados ? "S" : "N");
  
  // Grava o tamanho do corpo do robo
  for (j = 0; j < NUM_DIMENSOES; j++)
    fprintf(fp, "%.2f ", vdTamCorpoRobot[j]);
  fprintf(fp, "\n");
 
  // Grava o tamanho das partes das pernas  
  for (i = 0; i < iNumPartesPorPerna; i++) {
    for (j = 0; j < NUM_DIMENSOES; j++)
      fprintf(fp, "%.2f ", ppdTamPartesPernas[i][j]);
  }
  fprintf(fp, "\n");
  
  // Grava as restricoes das juntas
  for (i = 0; i < iNumPartesPorPerna; i++) {
    for (j = 0; j < 2; j++)
      fprintf(fp, "%.8f ", ppdRestricoesJunta[i][j]);
  }
  fprintf(fp, "\n");
  fclose(fp);
}


void AlocarMemoriaAutomato()
{
  int i;

  // Aloca memoria para o automato
  pdVelocRelativa = new double[iNumEstados + 1];
  for (i = 0; i <= iNumEstados; i++)
    pdVelocRelativa[i] = 0.0;
  ppdAngleAutomato = new double*[iTotalPartesPernas];
  ppdVelocAutomato = new double*[iTotalPartesPernas];
  for (i = 0; i < iTotalPartesPernas; i++) {
    ppdAngleAutomato[i] = new double[iNumEstados + 1];
    ppdVelocAutomato[i] = new double[iNumEstados + 1];
  }
}


void LiberarMemoriaRobotnik()
{
  int i;

  // Libera a memoria do robot e do automato
  if (ppdTamPartesPernas != NULL) {
    for (i = 0; i < iTotalPartesPernas; i++) {
      if (ppdTamPartesPernas[i] != NULL)
        delete [] ppdTamPartesPernas[i];
    }
    delete [] ppdTamPartesPernas;
    ppdTamPartesPernas = NULL;
  }
  if (ppdPosPartesPernas != NULL) {
    for (i = 0; i < iTotalPartesPernas; i++) {
      if (ppdPosPartesPernas[i] != NULL)
        delete [] ppdPosPartesPernas[i];
    }
    delete [] ppdPosPartesPernas;
    ppdPosPartesPernas = NULL;
  }
  if (ppdPosJuntasPernas != NULL) {
    for (i = 0; i < iTotalPartesPernas; i++) {
      if (ppdPosJuntasPernas[i] != NULL)
        delete [] ppdPosJuntasPernas[i];
    }
    delete [] ppdPosJuntasPernas;
    ppdPosJuntasPernas = NULL;
  }
  if (ppdRestricoesJunta != NULL) {
    for (i = 0; i < iTotalPartesPernas; i++) {
      if (ppdRestricoesJunta[i] != NULL)
        delete [] ppdRestricoesJunta[i];
    }
    delete [] ppdRestricoesJunta;
    ppdRestricoesJunta = NULL;
  }
  if (pbContatoComSolo != NULL)
    delete [] pbContatoComSolo;
  pbContatoComSolo = NULL;
  if (pbBumperSamples != NULL) {
    for (i = 0; i < MAX_SAMPLES_BUMPERS; i++)
      if (pbBumperSamples[i] != NULL)
        delete [] pbBumperSamples[i];
    delete [] pbBumperSamples;
  }
}


void LiberarMemoriaAutomato()
{
  int i;

  // Libera a memoria alocada pelo automato
  if (ppdAngleAutomato != NULL) {
    for (i = 0; i < iTotalPartesPernas; i++) {
      if (ppdAngleAutomato[i] != NULL)
        delete [] ppdAngleAutomato[i];
    }
    delete [] ppdAngleAutomato;
    ppdAngleAutomato = NULL;
  }
  if (ppdVelocAutomato != NULL) {
    for (i = 0; i < iTotalPartesPernas; i++) {
      if (ppdVelocAutomato[i] != NULL)
        delete [] ppdVelocAutomato[i];
    }
    delete [] ppdVelocAutomato;
    ppdVelocAutomato = NULL;
  }
  if (pdVelocRelativa != NULL)
    delete [] pdVelocRelativa;
  pdVelocRelativa = NULL;
}


/*---------------------------------------------------------------------------------------------------------------------*
 * Mundo fisico e colisivo                                                                                             *
 *---------------------------------------------------------------------------------------------------------------------*/
void CriaUniverso()
{
  if (bSimulacaoVisual) {
    // Seta os parametros da funcao de fisualizacao
    myDrawFunction.version = DS_VERSION;
    myDrawFunction.start = &DrawstuffStart;
    myDrawFunction.step = &DrawstuffSimLoop;
    myDrawFunction.command = &DrawstuffCommand;
    myDrawFunction.path_to_textures = PATH_TEXTURES;
  }

  // Criacao e inicializacao do mundo fisico
  World = dWorldCreate();
  dWorldSetGravity(World, 0.0, 0.0, GRAVITY);
  dWorldSetERP(World, ERP);
  dWorldSetCFM(World, CFM);

  // Criacao do espaco de colisao, do grupo de contatos e do solo
  Space = dSimpleSpaceCreate(0);
  geomSolo = dCreatePlane(Space, 0.0, 0.0, 1.0, 0.0);
}


void CriaTodosObjetos()
{
  int i;
  dMass massObjects;
  dMatrix3 rotacaoMatrix;

  // ********** Cria os membros ********** //
  bodyMembros = new dBodyID[iTotalPartesPernas];
  geomMembros = new dGeomID[iTotalPartesPernas];
  for (i = 0; i < iTotalPartesPernas; i++) {
    // Parte fisica
    bodyMembros[i] = dBodyCreate(World);
    dMassSetBox(&massObjects, DENSIDADE, ppdTamPartesPernas[i][0], ppdTamPartesPernas[i][1], ppdTamPartesPernas[i][2]);
    dBodySetMass(bodyMembros[i], &massObjects);
    dBodySetPosition(bodyMembros[i], ppdPosPartesPernas[i][0], ppdPosPartesPernas[i][1], ppdPosPartesPernas[i][2]);
    // Parte colisiva
    geomMembros[i] = dCreateBox(Space, ppdTamPartesPernas[i][0], ppdTamPartesPernas[i][1], ppdTamPartesPernas[i][2]);
    dGeomSetPosition(geomMembros[i], ppdPosPartesPernas[i][0], ppdPosPartesPernas[i][1], ppdPosPartesPernas[i][2]);
    // Conecta a parte fisica com a colisiva
    dGeomSetBody(geomMembros[i], bodyMembros[i]);
  }

  // ********** Cria o corpo ********** //
  // Parte fisica
  bodyCorpo = dBodyCreate(World);
  dMassSetBox(&massObjects, DENSIDADE, vdTamCorpoRobot[0], vdTamCorpoRobot[1], vdTamCorpoRobot[2]);
  dBodySetMass(bodyCorpo, &massObjects);
  dBodySetPosition(bodyCorpo, vdPosCorpoRobot[0], vdPosCorpoRobot[1], vdPosCorpoRobot[2]);
  // Parte colisiva
  geomCorpo = dCreateBox(Space, vdTamCorpoRobot[0], vdTamCorpoRobot[1], vdTamCorpoRobot[2]);
  dGeomSetPosition(geomCorpo, vdPosCorpoRobot[0], vdPosCorpoRobot[1], vdPosCorpoRobot[2]);
  // Conecta a parte fisica com a colisiva
  dGeomSetBody(geomCorpo, bodyCorpo);

  // ********** Cria as juntas visuais ********** //
  if (bSimulacaoVisual) {
    dRFromAxisAndAngle(rotacaoMatrix, 1, 0, 0, NOVENTA_GRAUS_RAD);
    bodyJuntas = new dBodyID[iTotalPartesPernas];
    for (i = 0; i < iTotalPartesPernas; i++) {
      bodyJuntas[i] = dBodyCreate(World);
      dMassSetCylinder(&massObjects, DENSIDADE / 1000.0, 2, Menor(ppdTamPartesPernas[i][1], (i+1) % iNumPartesPorPerna ? 
          ppdTamPartesPernas[i+1][1] : vdTamCorpoRobot[1]), Menor(Menor(ppdTamPartesPernas[i][0]/2.0, ((i+1) % 
          iNumPartesPorPerna ? ppdTamPartesPernas[i+1][0] : vdTamCorpoRobot[0])/2.0), Menor(ppdTamPartesPernas[i][2]/2.0, 
          ((i+1) % iNumPartesPorPerna ? ppdTamPartesPernas[i+1][2] : vdTamCorpoRobot[2])/2.0)));
      dBodySetMass(bodyJuntas[i], &massObjects);
      dBodySetPosition(bodyJuntas[i], ppdPosJuntasPernas[i][0], ppdPosJuntasPernas[i][1], ppdPosJuntasPernas[i][2]);
      dBodySetRotation(bodyJuntas[i], rotacaoMatrix);
    }
  }
}


void CriaJuntas()
{
  int i;

  // Cria as juntas e motores
  hingeMembros = new dJointID[iTotalPartesPernas];
  motorMembros = new dJointID[iTotalPartesPernas];
  fixedMembros = new dJointID[iTotalPartesPernas];

  for (i = 0; i < iTotalPartesPernas; i++) {
    // Cria as juntas
    hingeMembros[i] = dJointCreateHinge(World, 0);
    dJointAttach(hingeMembros[i], bodyMembros[i], (((i + 1) % iNumPartesPorPerna) ? bodyMembros[i + 1] : bodyCorpo));
    dJointSetHingeAxis(hingeMembros[i], 0.0, 1.0, 0.0);
    dJointSetHingeAnchor(hingeMembros[i], ppdPosJuntasPernas[i][0], ppdPosJuntasPernas[i][1], ppdPosJuntasPernas[i][2]);
    dJointSetHingeParam(hingeMembros[i], dParamBounce, 0.0);
    dJointSetHingeParam(hingeMembros[i], dParamFMax, 0.0);
    // O ponto zero tem que estar dentro da faixa de restricoes (senao o modelo explode)
    if (ppdRestricoesJunta[i][0] > 0.0)
      dJointSetHingeParam(hingeMembros[i], dParamLoStop, 0.0);
    else
      dJointSetHingeParam(hingeMembros[i], dParamLoStop, ppdRestricoesJunta[i][0]);
    if (ppdRestricoesJunta[i][1] < 0.0)
      dJointSetHingeParam(hingeMembros[i], dParamHiStop, 0.0);
    else
      dJointSetHingeParam(hingeMembros[i], dParamHiStop, ppdRestricoesJunta[i][1]);

    // Cria os motores das juntas
    motorMembros[i] = dJointCreateAMotor(World, 0);
    dJointSetAMotorMode(motorMembros[i], dAMotorUser);
    dJointAttach(motorMembros[i], bodyMembros[i], (((i + 1) % iNumPartesPorPerna) ? bodyMembros[i + 1] : bodyCorpo));
    dJointSetAMotorNumAxes(motorMembros[i], 1);
    dJointSetAMotorAxis(motorMembros[i], 0, 0, 0.0, 1.0, 0.0);
    dJointSetAMotorParam(motorMembros[i], dParamFMax, FORCA_MAX);
    dJointSetAMotorParam(motorMembros[i], dParamVel, 0.0);

    // Cria cilindros fisicos nas juntas com densidade pequena
    if (bSimulacaoVisual) {
      fixedMembros[i] = dJointCreateFixed(World, 0);
      dJointAttach(fixedMembros[i], bodyJuntas[i], bodyMembros[i]);
      dJointSetFixed(fixedMembros[i]);
    }
  }
}


void DestruirUniverso()
{ 
  // Destroi o universo  
  if (Space != NULL)
    dSpaceDestroy(Space);
  Space = NULL;
  
  if (World != NULL)
    dWorldDestroy(World);
  World = NULL;
  
  // Desaloca memoria
  if (fixedMembros != NULL)
    delete [] fixedMembros;
  fixedMembros = NULL;
  
  if (motorMembros != NULL)
    delete [] motorMembros;
  motorMembros = NULL;
  
  if (hingeMembros != NULL)
    delete [] hingeMembros;
  hingeMembros = NULL;
  
  if (bodyJuntas != NULL)
    delete [] bodyJuntas;
  bodyJuntas = NULL;

  if (geomMembros != NULL)
    delete [] geomMembros;
  geomMembros = NULL;
    
  if (bodyMembros != NULL)
    delete [] bodyMembros;
  bodyMembros = NULL;
}


/*---------------------------------------------------------------------------------------------------------------------*
 * Funcoes de controle da simulacao                                                                                    *
 *---------------------------------------------------------------------------------------------------------------------*/
void SetaPosOriginal()
{
  int i;
  dQuaternion quatObj;
  dMatrix3 rotacaoMatrix;

  // Restaura a posicao original e atitude dos objetos fisicos
  dQSetIdentity(quatObj);
  for (i = 0; i < iTotalPartesPernas; i++) {
    dBodySetPosition(bodyMembros[i], ppdPosPartesPernas[i][0], ppdPosPartesPernas[i][1], ppdPosPartesPernas[i][2]);
    dBodySetQuaternion(bodyMembros[i], quatObj);
    dGeomSetPosition(geomMembros[i], ppdPosPartesPernas[i][0], ppdPosPartesPernas[i][1], ppdPosPartesPernas[i][2]);
    dGeomSetQuaternion(geomMembros[i], quatObj);
  }
  dBodySetPosition(bodyCorpo, vdPosCorpoRobot[0], vdPosCorpoRobot[1], vdPosCorpoRobot[2]);
  dBodySetQuaternion(bodyCorpo, quatObj);
  dGeomSetPosition(geomCorpo, vdPosCorpoRobot[0], vdPosCorpoRobot[1], vdPosCorpoRobot[2]);
  dGeomSetQuaternion(geomCorpo, quatObj);

  if (bSimulacaoVisual) {
    dRFromAxisAndAngle(rotacaoMatrix, 1, 0, 0, NOVENTA_GRAUS_RAD);
    for (i = 0; i < iTotalPartesPernas; i++) {
      dBodySetPosition(bodyJuntas[i], ppdPosJuntasPernas[i][0], ppdPosJuntasPernas[i][1], ppdPosJuntasPernas[i][2]);
      dBodySetRotation(bodyJuntas[i], rotacaoMatrix);
    }
  }
}


double PhysicalSimLoop(const double dTimeSimulation)
{
  const dReal *corpoPosAtual = dBodyGetPosition(bodyCorpo);
  double dPosAntX = corpoPosAtual[0], dPosAntY = corpoPosAtual[1], dPosAntZ = corpoPosAtual[2];
  double dRelogioSimulacao = 0.0, dTimeAntGiro = 0.0;
  register bool bRobotnikCaindo = true;
  register int i;


  // Inicializacoes dos sensores
  iCountBumperSamples = iCountGiroSamples = 0;
  for (i = 0; i < iNumPernasRobotnik; i++)
    pbContatoComSolo[i] = false;
  iEstadoAtual = 0;

  // Laco de simulacao fisico
  while (dRelogioSimulacao < dTimeSimulation) {
    AjustarVelocidades();
    dSpaceCollide(Space, 0, NearCallback);
    dWorldStep(World, PASSO_FISICO);
    dJointGroupEmpty(ContactGroup);
    dRelogioSimulacao += PASSO_FISICO;

    // Leitura do giroscopio
    if (dRelogioSimulacao - dTimeAntGiro > SAMPLE_GYRO_INTERVAL && iCountGiroSamples < MAX_SAMPLES_GYRO) {
      corpoPosAtual = dBodyGetPosition(bodyCorpo);
      vdGyroSamplesX[iCountGiroSamples] = (corpoPosAtual[0] - dPosAntX);
      vdGyroSamplesY[iCountGiroSamples] = (corpoPosAtual[1] - dPosAntY);
      vdGyroSamplesZ[iCountGiroSamples] = (corpoPosAtual[2] - dPosAntZ);
      dPosAntX = corpoPosAtual[0];
      dPosAntY = corpoPosAtual[1];
      dPosAntZ = corpoPosAtual[2];
      dTimeAntGiro = dRelogioSimulacao;
      iCountGiroSamples++;
    }

    // Leitura dos bumpers
    bRobotnikCaindo = true;
    if (iCountBumperSamples < MAX_SAMPLES_BUMPERS) {
      for (i = 0; i < iNumPernasRobotnik; i++) {
        pbBumperSamples[iCountBumperSamples][i] = pbContatoComSolo[i] && AlturaPataOk(i);
        if (bRobotnikCaindo && pbBumperSamples[iCountBumperSamples][i])
          bRobotnikCaindo = false;
        pbContatoComSolo[i] = false;
      }
      iCountBumperSamples++;
      if (bRobotnikCaindo)
        break;
    }
  }
  PararTodasAsJuntas();
  return dRelogioSimulacao;
}


void PararTodasAsJuntas()
{
  int i;

  // Para todas as juntas do robo
  for (i = 0; i < iTotalPartesPernas; i++)
    dJointSetAMotorParam(motorMembros[i], dParamVel, 0.0);
  iEstadoAtual = -1;
}


inline void AjustarVelocidades()
{
  static double dTimeAcumulado = 0.0;
  register int i, iNumParados = 0;
  double dAngleAtual;

  if (iEstadoAtual < 0) {
    dTimeAcumulado = 0.0;
    return;
  }

  // Calcula as velocidades das juntas e as aplica
  for (i = 0; i < iTotalPartesPernas; i++) {
    // Verifica se a junta ja atingiu o estado otimo
    dAngleAtual = dJointGetHingeAngle(hingeMembros[i]);
    if (((ppdVelocAutomato[i][iEstadoAtual] < 0) && (dAngleAtual > ppdAngleAutomato[i][iEstadoAtual]))
        || ((ppdVelocAutomato[i][iEstadoAtual] > 0) && (dAngleAtual < ppdAngleAutomato[i][iEstadoAtual])))
      dJointSetAMotorParam(motorMembros[i], dParamVel, ppdVelocAutomato[i][iEstadoAtual]);
    else {
      dJointSetAMotorParam(motorMembros[i], dParamVel, 0.0);
      iNumParados++;
    }
  }

  // Verifica a troca de estados - todos segmentos parados ou limite de tempo
  dTimeAcumulado += PASSO_FISICO;
  if (iNumParados == iTotalPartesPernas || (dTimeAcumulado > TEMPO_MAX_ESTADO / pdVelocRelativa[iEstadoAtual] && iEstadoAtual > 0)) {
    iEstadoAtual = iEstadoAtual % iNumEstados + 1;
    dTimeAcumulado = 0.0;
  }
}


void SetAutomatoFromFile(const string sNomeArquivo)
{
  FILE *fp = NULL;
  int j, iPernas, iPartes, i = 0;
  char vcLinha[MAX_LINHA + 1];
  char *szPalavra;

  // Abre o arquivo
  if ((fp = fopen(sNomeArquivo.c_str(), "r")) == NULL) {
    printf("Nao foi possivel abrir o arquivo do automato\n");
    exit(1);
  }

  // Busca o numero de pernas, de partes por perna e de estados na primeira linha
  fgets(vcLinha, MAX_LINHA, fp);
  if ((szPalavra = strtok(vcLinha, " ")) == NULL) {
    printf("Arquivo do automato invalido\n");
    exit(1);
  }
  iPernas = atoi(szPalavra);
  szPalavra = strtok('\0', " ");
  iPartes = atoi(szPalavra);
  szPalavra = strtok('\0', " ");
  iNumEstados = atoi(szPalavra);
  szPalavra = strtok('\0', " ");
  if (iPernas != iNumPernasRobotnik || iPartes != iNumPartesPorPerna || !iNumEstados || !iNumEstados % 2) {
    printf("Arquivo do automato com dimensoes incorretas\n");
    exit(1);
  }
  AlocarMemoriaAutomato();

  // Busca as velocidades relativas de cada estado
  fgets(vcLinha, MAX_LINHA, fp);
  szPalavra = strtok(vcLinha, " ");
  for (i = 0; i <= iNumEstados && szPalavra != NULL; i++) {
    pdVelocRelativa[i] = atof(szPalavra);
    szPalavra = strtok('\0', " ");
  }

  // Le as linhas para preencher a tabela de angulos
  i = 0;
  while (!feof(fp) && i < iTotalPartesPernas) {
    if (fgets(vcLinha, MAX_LINHA, fp) && strlen(vcLinha) > 0) {
      j = 1;
      szPalavra = strtok(vcLinha, " ");
      while (szPalavra != NULL && j <= iNumEstados) {
        ppdAngleAutomato[i][j++] = atof(szPalavra);
        szPalavra = strtok('\0', " ");
      }
      i++;
    }
  }
  fclose(fp);
  CalcularVelocidades();
}


void CalcularVelocidades()
{
  int i, j;

  // Preenche a tabela de velocidades
  for (i = 0; i < iTotalPartesPernas; i++) {
    // Preenche o estado inicial
    ppdAngleAutomato[i][0] = ppdAngleAutomato[i][iNumEstados];
    ppdVelocAutomato[i][0] = ppdAngleAutomato[i][0] * pdVelocRelativa[0];
    // Preenche os demais estados
    for (j = 1; j <= iNumEstados; j++)
      ppdVelocAutomato[i][j] = (ppdAngleAutomato[i][j] - ppdAngleAutomato[i][j - 1]) * pdVelocRelativa[j];
  }
}


void AjustarAngulosPatas()
{
  double dAngulo;
  int i, j, k;

  // Calcula os pes automaticamente pela cinematica direta
  for (j = 1; j <= iNumEstados; j++) {
    for (k = 0; k < iNumPernasRobotnik; k++) {
      // Soma os angulos das outras juntas
      dAngulo = 0.0;
      for (i = 1; i < iNumPartesPorPerna; i++)
        dAngulo += ppdAngleAutomato[k * iNumPartesPorPerna + i][j];
      // Aplica a soma inversa para que o pe fique paralelo ao chao
      ppdAngleAutomato[k * iNumPartesPorPerna][j] = -dAngulo;
    }
  }
  // Verifica restricoes depois disto
  VerificarRestricoes();
}


void VerificarRestricoes()
{
  int i, j;

  // Verifica se nao excede as restricoes
  for (i = 0; i < iTotalPartesPernas; i++) {
    for (j = 1; j <= iNumEstados; j++) {
      if (ppdAngleAutomato[i][j] < ppdRestricoesJunta[i][0])
        ppdAngleAutomato[i][j] = ppdRestricoesJunta[i][0];
      else if (ppdAngleAutomato[i][j] > ppdRestricoesJunta[i][1])
        ppdAngleAutomato[i][j] = ppdRestricoesJunta[i][1];
    }
  }
}


/*---------------------------------------------------------------------------------------------------------------------*
 * Funcoes da cinematica inversa                                                                                       *
 *---------------------------------------------------------------------------------------------------------------------*/
void AlocarMemoriaPowell()
{
  int i;

  // Cria o vetor a ser otimizado
  pdAnglesCinemat = new double[iNumPartesPorPerna + 1];
  for (i = 0; i <= iNumPartesPorPerna; i++)
    pdAnglesCinemat[i] = 0.0;

  // Cria a matriz identidade
  ppdMatrixIdenty = new double*[iNumPartesPorPerna + 1];
  for (i = 0; i <= iNumPartesPorPerna; i++)
    ppdMatrixIdenty[i] = new double[iNumPartesPorPerna + 1];
  InicializarMatrizIdentidade();

  // Aloca memoria para os recipes
  AlocarMemoriaRecipes(iNumPartesPorPerna + 1);
}


inline void InicializarMatrizIdentidade()
{
  register int i, j;

  // Inicializa a matriz identidade com a diagonal principal em 1 e o resto 0
  for (i = 0; i <= iNumPartesPorPerna; i++) {
    for (j = 0; j <= iNumPartesPorPerna; j++) {
      if (i == j)
        ppdMatrixIdenty[i][j] = 1.0;
      else
        ppdMatrixIdenty[i][j] = 0.0;
    }
  }
}


static double CinematicaInversa(double dAngles[])
{
  register int i, iDesloc, iParte;
  double dRestrX, dRestrY, dRestrA = 0.0, dAngulo = 0.0, dValorX = 0.0, dValorY = 0.0, dRestrJ = 0.0;

  // Calcula a cinematica direta
  iParte = iPernaAtualCinemat * iNumPartesPorPerna + iNumPartesPorPerna;
  for (i = 1; i <= iNumPartesPorPerna; i++) {
    dAngulo += dAngles[i];
    dValorX += cos(dAngulo) * ppdTamPartesPernas[iParte - i][2];
    dValorY += sin(dAngulo) * ppdTamPartesPernas[iParte - i][2];
  }
  dRestrX = dValorX - dPosDesejadaX;
  dRestrY = dValorY - dPosDesejadaY;

  // Restricao do angulo das patas
  if (bAngulosPatasCalculados)
    dRestrA = Quadrado(dAngulo + NOVENTA_GRAUS_RAD);

  // Restricao das juntas
  iDesloc = iPernaAtualCinemat * iNumPartesPorPerna;
  iParte = iDesloc + iNumPartesPorPerna - 1;
  if (dAngles[1] < ppdRestricoesJunta[iParte][0] - NOVENTA_GRAUS_RAD)
    dRestrJ += Quadrado(dAngles[1] - ppdRestricoesJunta[iParte][0] - NOVENTA_GRAUS_RAD);
  else if (dAngles[1] > ppdRestricoesJunta[iParte][1] - NOVENTA_GRAUS_RAD)
    dRestrJ += Quadrado(dAngles[1] - ppdRestricoesJunta[iParte][1] - NOVENTA_GRAUS_RAD);
  for (i = 2; i <= iNumPartesPorPerna; i++) {
    iParte = iDesloc + iNumPartesPorPerna - i;
    if (dAngles[i] < ppdRestricoesJunta[iParte][0])
      dRestrJ += Quadrado(dAngles[i] - ppdRestricoesJunta[iParte][0]);
    else if (dAngles[i] > ppdRestricoesJunta[iParte][1])
      dRestrJ += Quadrado(dAngles[i] - ppdRestricoesJunta[iParte][1]);
  }
  return Quadrado(dRestrX) + Quadrado(dRestrY) + dRestrJ + dRestrA;
}


inline double SimplificarAngulo(const double dAngle)
{
  const double RAD360G = 6.283185307179586;
  double dNewAngle = dAngle;
  register int iNumVoltas;

  // Simplifica os angulos de forma que eles fiquem entre -180 e 180
  if (dAngle > RAD360G || dAngle < -RAD360G) {
    iNumVoltas = (int) (dAngle / RAD360G);
    dNewAngle = dAngle - RAD360G * (double) iNumVoltas;
  }
  if (dNewAngle > M_PI)
    dNewAngle = -M_PI + dNewAngle;
  else if (dNewAngle < -M_PI)
    dNewAngle = dNewAngle + M_PI;
  return dNewAngle;
}


double CalcularTamanhoPerna(const int iPerna)
{
  int i;
  double dTamanho = 0.0;

  // Retorna o tamanho da perna iPerna
  for (i = 0; i < iNumPartesPorPerna; i++)
    dTamanho += ppdTamPartesPernas[iPerna * iNumPartesPorPerna + i][2];
  return dTamanho;
}


void AjustaDistanciaPonto(double &dPosX, double &dPosY)
{
  double dAngulo, dTamanhoPerna = CalcularTamanhoPerna(iPernaAtualCinemat);
  double dDistancia = sqrt(Quadrado(dPosX) + Quadrado(dPosY));

  // Se a distancia for maior que o tamanho da perna, traz o ponto para perto mantendo o angulo
  if (dDistancia > dTamanhoPerna)
  {
    dAngulo = atan(fabs(dPosX / dPosY));
    dPosX = sin(dAngulo) * dTamanhoPerna * Sinal(dPosX);
    dPosY = cos(dAngulo) * dTamanhoPerna * Sinal(dPosY);
  }
}


bool SetTudinhoFromGenoma(GAGenome & g)
{
  GARealGenome & genome = (GARealGenome &) g;
  register int i, j, k, o;
  int iNumeroIteracoes, iCount = 0;
  double dRetornoCinemat, dTamanho;


  // Preenche as velocidades relativas
  pdVelocRelativa[0] = (double) genome.gene(iCount++);
  for (j = 1; j <= iNumEstados / 2; j++) {
    o = (j + iNumEstados / 2 - 1) % iNumEstados + 1;
    pdVelocRelativa[j] = pdVelocRelativa[o] = (double) genome.gene(iCount++);
  }

  // Monta a tabela de angulos atraves da cinematica inversa
  iPernaAtualCinemat = 0;
  for (j = 1; j <= iNumEstados; j++) {
    dPosDesejadaX = (double) genome.gene(iCount++);
    dPosDesejadaY = (double) genome.gene(iCount++);
    AjustaDistanciaPonto(dPosDesejadaX, dPosDesejadaY);

    for (k = 0; k < iNumPernasRobotnik; k++) {
      // Informa qual e a perna atual - a funcao CinematicaInversa precisa saber
      iPernaAtualCinemat = k;

      // Se as pernas sao sincronizadas, e preciso setar os angulos em estados opostos para pernas opostas
      if ((iNumPernasRobotnik == 4 && (k == 0 || k == 3)) || (iNumPernasRobotnik == 6 && !(k % 2)))
        o = (j + iNumEstados / 2 - 1) % iNumEstados + 1;
      else
        o = j;

      // Seta os angulos do estado anterior desta perna para facilidar a busca
      if (o > 1) {
        for (i = 0; i < iNumPartesPorPerna; i++) {
          if (iNumPartesPorPerna - i == 1)
            pdAnglesCinemat[iNumPartesPorPerna - i] = SimplificarAngulo(ppdAngleAutomato[i + k *
                iNumPartesPorPerna][o - 1] - NOVENTA_GRAUS_RAD);
          else
            pdAnglesCinemat[iNumPartesPorPerna - i] = ppdAngleAutomato[i + k * iNumPartesPorPerna][o - 1];
        }
      }
      else {
        for (i = 0; i <= iNumPartesPorPerna; i++)
          pdAnglesCinemat[i] = 0.0;
        pdAnglesCinemat[1] = -NOVENTA_GRAUS_RAD;
      }

      // Calcula a cinematica inversa
      InicializarMatrizIdentidade();
      powell(pdAnglesCinemat, ppdMatrixIdenty, iNumPartesPorPerna, TOLERANCIA_POWELL, &iNumeroIteracoes,
          &dRetornoCinemat, CinematicaInversa);

      // Grava os angulos na tabela
      for (i = 0; i < iNumPartesPorPerna; i++) {
        // Acrescenta 90 no da coxa para girar a perna para baixo
        if (iNumPartesPorPerna - i == 1)
          ppdAngleAutomato[i + k * iNumPartesPorPerna][o] = SimplificarAngulo
              (pdAnglesCinemat[iNumPartesPorPerna - i] + NOVENTA_GRAUS_RAD);
        else
          ppdAngleAutomato[i + k * iNumPartesPorPerna][o] = SimplificarAngulo(pdAnglesCinemat[iNumPartesPorPerna - i]);
      }
    }
  }

  // Verifica  as restricoes, o alinhamento das patas e calcula as velocidades
  VerificarRestricoes();
  if (bAngulosPatasCalculados)
    AjustarAngulosPatas();
  CalcularVelocidades();
  
  
  // Comprimento e largura do corpo
  for (k = 0; k < NUM_DIMENSOES; k++)
    vdTamCorpoRobot[k] = (double) genome.gene(iCount++);
  
  // Pega o tamanho das partes das pernas do robotnik
  for (j = 0; j < iNumPartesPorPerna; j++) {
    for (k = 0; k < NUM_DIMENSOES; k++) {
      dTamanho = (double) genome.gene(iCount++);
      for (i = 0; i < iNumPernasRobotnik; i++)
        ppdTamPartesPernas[i * iNumPartesPorPerna + j][k] = dTamanho;
    }
  }
  
  // Valida os tamanhos para o modelo nao estourar
  for (j = 0; j < iNumPartesPorPerna; j++) {
    if ((ppdTamPartesPernas[j][0] >= vdTamCorpoRobot[0] / 3.0) || (ppdTamPartesPernas[j][1] >= vdTamCorpoRobot[1] / 2.0))
      return false;
  }

  // Calcula as posicoes das partes do robotnik
  CalcularPosicoesRobotnik();
  return true;
}


void LiberarMemoriaPowell()
{
  int i;

  if (ppdMatrixIdenty != NULL) {
    for (i = 0; i <= iNumPartesPorPerna; i++) {
      if (ppdMatrixIdenty[i] != NULL)
        delete [] ppdMatrixIdenty[i];
    }
    delete [] ppdMatrixIdenty;
    ppdMatrixIdenty = NULL;
  }
  if (pdAnglesCinemat != NULL)
    delete [] pdAnglesCinemat;
  pdAnglesCinemat = NULL;

  // Libera a memoria dos recipes
  LiberarMemoriaRecipes();
}


void GravarAutomato(const string sNomeArquivo)
{
  FILE *fp = NULL;
  int i, j;

  // Salva o automato atual em um arquivo
  if ((fp = fopen(sNomeArquivo.c_str(), "w")) == NULL) {
    printf("Nao foi possivel criar o arquivo do automato \'%s\'\n", sNomeArquivo.c_str());
    exit(1);
  }

  // Grava o numero de pernas, juntas e de estados
  fprintf(fp, "%d %d %d\n", iNumPernasRobotnik, iNumPartesPorPerna, iNumEstados);

  // Grava as velocidades relativas
  for (i = 0; i <= iNumEstados; i++)
    fprintf(fp, "%10.6f ", pdVelocRelativa[i]);
  fprintf(fp, "\n");

  // Grava os angulos
  for (i = 0; i < iTotalPartesPernas; i++) {
    for (j = 1; j <= iNumEstados; j++)
      fprintf(fp, "%10.6f ", ppdAngleAutomato[i][j]);
    fprintf(fp, "\n");
  }
  fclose(fp);
}


/*---------------------------------------------------------------------------------------------------------------------*
 * Funcoes sensoriais                                                                                                  *
 *---------------------------------------------------------------------------------------------------------------------*/
double GetAlturaMaxBumpers()
{
  const double dInclinacaoMax = DegToRad(7.5);
  double dAngulo, dAlturaMax;

  // Calcula a altura maxima para o pe ser considerado no chao
  dAngulo = atan(ppdTamPartesPernas[0][2] / ppdTamPartesPernas[0][0]);
  if (ppdTamPartesPernas[0][2] > ppdTamPartesPernas[0][0]) // Pes mais altos do que largos (cavalo)
    dAlturaMax = (ppdTamPartesPernas[0][2] / 2.0) / sin(dAngulo);
  else  // Pes mais largos do que altos (humanos)
    dAlturaMax = sin(dInclinacaoMax + dAngulo) * ((ppdTamPartesPernas[0][0] / 2.0) / cos(dAngulo));
  return dAlturaMax;
}


double GetAlturaMinBumpers()
{
  const double dInclinacaoMax = DegToRad(15.0);
  double dAngulo, dAlturaMin = 0.0;
  double dAlturaMax = GetAlturaMaxBumpers();

  // Calcula a altura minima para o pe ser considerado no chao
  dAngulo = atan(ppdTamPartesPernas[0][0] / ppdTamPartesPernas[0][2]);
  if (ppdTamPartesPernas[0][2] > ppdTamPartesPernas[0][0])
    dAlturaMin = sin(dInclinacaoMax + dAngulo) * dAlturaMax;
  return dAlturaMin;
}


inline bool AlturaPataOk(const int iNumeroPata)
{
  static double dAlturaMax = GetAlturaMaxBumpers();
  static double dAlturaMin = GetAlturaMinBumpers();
  double dAlturaAtual;

  // Conta o numero de pes que estao em contato com o chao
  dAlturaAtual = dBodyGetPosition(bodyMembros[iNumeroPata * iNumPartesPorPerna])[2];
  if (dAlturaAtual <= dAlturaMax && dAlturaAtual >= dAlturaMin)
    return true;
  else
    return false;
}


double CalcularInstabilidade()
{
  double dMediaX, dMediaY, dMediaZ, dVarianceX, dVarianceY, dVarianceZ;
  register int i;

  // Calcula a instabilidade do caminhar
  if (iCountGiroSamples < 2)
    return 1.0;

  // Calcula as medias
  dMediaX = dMediaY = dMediaZ = 0.0;
  for (i = 0; i < iCountGiroSamples; i++) {
    dMediaX += vdGyroSamplesX[i];
    dMediaY += vdGyroSamplesY[i];
    dMediaZ += vdGyroSamplesZ[i];
  }
  dMediaX /= (double) iCountGiroSamples;
  dMediaY /= (double) iCountGiroSamples;
  dMediaZ /= (double) iCountGiroSamples;

  // Calcula as variancias
  dVarianceX = dVarianceY = dVarianceZ = 0.0;
  for (i = 0; i < iCountGiroSamples; i++) {
    dVarianceX += Quadrado(vdGyroSamplesX[i] - dMediaX);
    dVarianceY += Quadrado(vdGyroSamplesY[i] - dMediaY);
    dVarianceZ += Quadrado(vdGyroSamplesZ[i] - dMediaZ);
  }

  // Retorna o desvio padrao das leituras do giroscopio
  return sqrt((dVarianceX + dVarianceY + dVarianceZ) / (iCountGiroSamples - 1.0));
}


double CalcularFatorBumpers()
{
  const int iMaxAmostras = (int) (dTimeSimulation / PASSO_FISICO);
  register int i, j, k, iCount;
  double dFactor = 0.0;

  if (iCountBumperSamples < iNumPernasRobotnik)
    return 1.0;

  // O fator varia de 1 (tudo no chao/ar) a 0 (trote ultra cadenciado)
  for (j = 0; j < iNumPernasRobotnik; j++) {
    iCount = 0;
    k = (j + iNumPernasRobotnik / 2) % iNumPernasRobotnik;
    for (i = 0; i < iCountBumperSamples; i++) {
      if (pbBumperSamples[i][j] && !pbBumperSamples[i][k])
        iCount++;
    }
    dFactor += Quadrado((double) iCount / (double) iCountBumperSamples - 0.5);
  }

  // Faz com que o fator fique entre 0 e 1 independente do nmero de pernas do robotnik
  dFactor = dFactor * 4.0 / (double) iNumPernasRobotnik;

  // Se a simulacao nao foi ate o fim, descontamos o fator
  if (iCountBumperSamples < iMaxAmostras)
    dFactor = dFactor * ((double) iCountBumperSamples / (double) iMaxAmostras) + ((double) iMaxAmostras - (double)
      iCountBumperSamples) / (double) iMaxAmostras;
  return dFactor;
}


double CalcularDeslocamento()
{
  double dPosIni = vdPosCorpoRobot[0];
  double dPosFim = dBodyGetPosition(bodyCorpo)[0];
  register int i;

  for (i = 0; i < iNumPernasRobotnik; i++) {
    dPosIni += ppdPosPartesPernas[i * iNumPartesPorPerna][0];
    dPosFim += dBodyGetPosition(bodyMembros[i * iNumPartesPorPerna])[0];
  }
  return (dPosFim - dPosIni) / ((double) iNumPernasRobotnik + 1.0);
}


/*---------------------------------------------------------------------------------------------------------------------*
 * Funcoes do Algoritmo Genetico                                                                                       *
 *---------------------------------------------------------------------------------------------------------------------*/
void CarregarParametros(const string sArqParams)
{
  // Busca os parametros geneticos
  lRandomSeed = atol(GetParamInFile(sArqParams, "random_seed").c_str());
  iNumEstados = atoi(GetParamInFile(sArqParams, "number_states").c_str());
  dTimeSimulation = atof(GetParamInFile(sArqParams, "time_simulation").c_str());
}


void AlgoritmoGenetico(const string sArqParams)
{
  long lTime, lTimeRestante;
  time_t tmAtual, tmInicial = time(NULL);
  double dTamanhoPerna = CalcularTamanhoPerna(0);
  int i, j;
  char szMensagem[MAX_LINHA + 1];
  lEpochOfBest = lGeracaoAtual = 0;

  // Cria os alelos
  GARealAlleleSetArray alleles;

  // Alelos para as velocidades relativas
  alleles.add(VELOC_RELATIVA_MIN, VELOC_RELATIVA_MAX / 2.0, INTERVALO_VELOCIDADE);
  for (j = 1; j <= iNumEstados / 2; j++)
    alleles.add(VELOC_RELATIVA_MIN, VELOC_RELATIVA_MAX, INTERVALO_VELOCIDADE);

  // Alelos para as posicoes
  for (j = 0; j < iNumEstados; j++) {
    alleles.add(-dTamanhoPerna * 0.50,  dTamanhoPerna * 0.5, INTERVALO_POSICAO);
    alleles.add(-dTamanhoPerna * 0.95, -dTamanhoPerna * 0.5, INTERVALO_POSICAO);
  }

   // Alelos para as dimensoes do corpo
  for (j = 0; j < NUM_DIMENSOES; j++)
    alleles.add(INTERVALO_ALELOS_MORFO, TAM_MAX_CORPO, INTERVALO_ALELOS_MORFO);
    
  // Alelos para as dimensoes das pernas
  for (i = 0; i < iNumPartesPorPerna; i++) {
    for (j = 0; j < NUM_DIMENSOES; j++)
      alleles.add(INTERVALO_ALELOS_MORFO, TAM_MAX_PARTES, INTERVALO_ALELOS_MORFO);
  }
  
  // Cria o genoma
  GARealGenome genome(alleles, Objective);
  genome.initializer(Initializer);
  genome.comparator(GARealGenome::ElementComparator);
  genome.mutator(GARealGaussianMutator);
  genome.crossover(GARealGenome::OnePointCrossover);

  // Cria e inicializa o GA
  GASteadyStateGA ga(genome);
  ga.maximize();
  ga.parameters(sArqParams.c_str());
  ga.recordDiversity(gaFalse);
  ga.selectScores(GAStatistics::Maximum | GAStatistics::Mean | GAStatistics::Deviation);
  GASigmaTruncationScaling sigmaTruncation;
  ga.scaling(sigmaTruncation);
  GASRSSelector srsSelector;
  ga.selector(srsSelector);

  // Informacoes iniciais
  printf("Tamanho da Population..:%5d\n", ga.populationSize());
  printf("Numero de Generations..:%5d\n", ga.nGenerations());
  printf("Taxa de Mutation.......:%10.4f\n", ga.pMutation());
  printf("Taxa de Crossover......:%10.4f\n", ga.pCrossover());
  printf("Taxa de Replacement....:%10.4f\n", ga.pReplacement());
  printf("Random seed utilizado..:%5li\n", lRandomSeed);
  printf("Frequencia relatorios..:%5d\n", ga.scoreFrequency());
  printf("Numero de estados......:%5d\n", iNumEstados);
  printf("Tempo de caminhada.....:%10.4f\n", dTimeSimulation);


  // Realiza a evolucao natural
  ga.initialize(lRandomSeed);
  while (!ga.done() && (ga.statistics().current(GAStatistics::Deviation) > 1.0e-4 || lGeracaoAtual < ga.nGenerations() / 5) 
      && !bAplicacaoCancelada && lGeracaoAtual - lEpochOfBest < ga.nGenerations() / 6) {
    // Passo evolutivo
    ga.step();
    lGeracaoAtual = ga.generation();

    // Exibe relatorio da evolucao
    if (!(lGeracaoAtual % ga.scoreFrequency())) {
      tmAtual = time(NULL);
      lTime = (long) (tmAtual - tmInicial);
      lTimeRestante = (long) ((double) ((tmAtual - tmInicial) / (lGeracaoAtual + 1.0)) * (ga.nGenerations() - lGeracaoAtual));
      sprintf(szMensagem, "GEN:%6li   MAX:%10.6f   MED:%10.6f   DEV:%10.6f   %02li:%02li:%02li  %02li:%02li:%02li\n",
          lGeracaoAtual, ga.statistics().current(GAStatistics::Maximum), ga.statistics().current(GAStatistics::Mean),
          ga.statistics().current(GAStatistics::Deviation), lTime / 3600, lTime % 3600 / 60, lTime % 60,
          lTimeRestante / 3600, lTimeRestante % 3600 / 60, lTimeRestante % 60);
      printf(szMensagem);
      if (bEcoTelaHabilitado) 
        fprintf(stderr, szMensagem);
    }
  }

  // Salva o melhor genoma no fim
  lTime = (long) (time(NULL) - tmInicial);
  printf("Tempo total de aprendizado: %02li:%02li:%02li\n", lTime / 3600, lTime % 3600 / 60, lTime % 60);
  ga.flushScores();
  genome = ga.statistics().bestIndividual();
  SetTudinhoFromGenoma(genome);
  GravarAutomato(sNomeArquivoBase + ".ctr");
  SalvarRobotnik(sNomeArquivoBase + ".rbt");
}


static void Initializer(GAGenome& g)
{
  GARealGenome & genome = (GARealGenome &) g;
  int i, j, iCount = 0;
  double vdTamCorpo[NUM_DIMENSOES];
  double vvdTamPartes[iNumPartesPorPerna][NUM_DIMENSOES];
  double dTamanhoPerna = CalcularTamanhoPerna(0);
  bool bTamanhoValido;

  
  // Alelos para as velocidades relativas
  genome.gene(iCount++, GARandomFloat(VELOC_RELATIVA_MIN, VELOC_RELATIVA_MAX / 2.0));  
  for (j = 1; j <= iNumEstados / 2; j++)
    genome.gene(iCount++, GARandomFloat(VELOC_RELATIVA_MIN, VELOC_RELATIVA_MAX));
  
  // Alelos para as posicoes
  for (j = 0; j < iNumEstados; j++) {
    genome.gene(iCount++, GARandomFloat(-dTamanhoPerna * 0.50,  dTamanhoPerna * 0.5));
    genome.gene(iCount++, GARandomFloat(-dTamanhoPerna * 0.95, -dTamanhoPerna * 0.5));
  }
  
  // Valida os tamanhos gerados antes de gravar nos alelos
  do {
    // Alelos para as dimensoes do corpo
    for (j = 0; j < NUM_DIMENSOES; j++)
      vdTamCorpo[j] = (double) GARandomFloat(INTERVALO_ALELOS_MORFO, TAM_MAX_CORPO);
     
    // Alelos para as dimensoes das pernas
    for (i = 0; i < iNumPartesPorPerna; i++) {
      for (j = 0; j < NUM_DIMENSOES; j++)
        vvdTamPartes[i][j] = GARandomFloat(INTERVALO_ALELOS_MORFO, TAM_MAX_PARTES);
    }  
    
    // Valida os tamanhos para o modelo nao estourar
    bTamanhoValido = true;
    for (j = 0; j < iNumPartesPorPerna; j++) {
      if ((vvdTamPartes[j][0] >= vdTamCorpo[0] / 3.0) || (vvdTamPartes[j][1] >= vdTamCorpo[1] / 2.0)) {
        bTamanhoValido = false;
        break;
      } 
    }  
  } while (!bTamanhoValido);
  

  // Preenche os alelos das dimensoes do corpo
  for (j = 0; j < NUM_DIMENSOES; j++)
    genome.gene(iCount++, vdTamCorpo[j]);
  
  // Preenche os alelos para as dimensoes das pernas
  for (i = 0; i < iNumPartesPorPerna; i++) {
    for (j = 0; j < NUM_DIMENSOES; j++)
      genome.gene(iCount++, vvdTamPartes[i][j]);
  }  
}


static float Objective(GAGenome & g)
{
  double dTempoSimulado, dFatorBumpers, dDeslocamento, dInstabilidade, dFitness;
  static double dMelhorFitness = 0.0;
  
  
  // Se o genoma for invalido, nem simula
  if (!SetTudinhoFromGenoma(g))
    return 0.0f;
    
  // Cria o robo com as especificacoes do genoma
  CriaUniverso();
  CriaTodosObjetos();
  CriaJuntas();

  // Executa a simulacao e calcula o fitness
  dTempoSimulado = PhysicalSimLoop(dTimeSimulation);
  dDeslocamento = CalcularDeslocamento();
  if (dDeslocamento > -5.0 * dTempoSimulado && dDeslocamento < 5.0 * dTempoSimulado) {
    dInstabilidade = CalcularInstabilidade();
    dFatorBumpers = CalcularFatorBumpers();
    dFitness = dDeslocamento / (1.0 + dFatorBumpers + PESO_GYRO * dInstabilidade);
  }
  else {
    dFitness = dDeslocamento = 0.0;
    dInstabilidade = dFatorBumpers = 1.0;
  }

  // Exibe os melhores fitness na tela e salva em um arquivo
  if (dFitness > dMelhorFitness) {
    dMelhorFitness = dFitness;
    lEpochOfBest = lGeracaoAtual;
    GravarAutomato(sNomeArquivoBase + ".ctr");
    SalvarRobotnik(sNomeArquivoBase + ".rbt");
    printf("FITN:%10.6f   DESL:%10.6f   BUMP:%9.6f   INST:%10.6f   TIME:%6.2f\n", dFitness, dDeslocamento,
        dFatorBumpers, dInstabilidade, dTempoSimulado);
  }
  // Seta a posicao original e retorna o fitness
  DestruirUniverso();
  return (float) dFitness;
}


/*---------------------------------------------------------------------------------------------------------------------*
 * Funcoes Utilitarias                                                                                                 *
 *---------------------------------------------------------------------------------------------------------------------*/
string GetParamInFile(const string sNameFile, const string sParam)
{
  FILE *fp = NULL;
  char vcLinha[MAX_LINHA + 1];
  char *szPalavra;

  // Busca parametros no arquivo de configuracao
  if ((fp = fopen(sNameFile.c_str(), "r")) == NULL) {
    printf("Nao foi possivel abrir o arquivo de parametros\n");
    exit(1);
  }
  while (!feof(fp)) {
    if (fgets(vcLinha, MAX_LINHA, fp) && strlen(vcLinha) > 0) {
      if (vcLinha[0] == '#')
        continue;
      szPalavra = strtok(vcLinha, " ");
      while (szPalavra != NULL) {
        if (!strcmp(szPalavra, sParam.c_str())) {
          if ((szPalavra = strtok('\0', " "))) {
            fclose(fp);
            return szPalavra;
          }
          else
            continue;
        }
        szPalavra = strtok('\0', " ");
      }
    }
  }
  fclose(fp);
  printf("Parametro nao encontrado no arquivo de parametros\n");
  exit(1);
}


inline double DegToRad(const double dValDeg)
{
  const double MULTIP = M_PI / 180.0;
  return dValDeg * MULTIP;
}


inline double RadToDeg(const double dValRad)
{
  const double MULTIP = 180.0 / M_PI;
  return dValRad * MULTIP;
}


inline double Quadrado(const double dValor)
{
  return dValor * dValor;
}


inline double Sinal(const double dValor)
{
  if (dValor < 0.0)
    return -1.0;
  else if (dValor > 0.0)
    return 1.0;
  else
    return 0.0;
}


inline double Menor(const double dValorA, const double dValorB)
{
  if (dValorA < dValorB)
    return dValorA;
  else
    return dValorB;
}


void HabilitaEcoTela(int iSinal) {
  if (iSinal == SIGUSR1) {
    fprintf(stderr, "Sinal SIGUSR1 recebido, eco em stderr habilitado\n");
    bEcoTelaHabilitado = true;
  }
  else if (iSinal == SIGUSR2) {
    fprintf(stderr, "Sinal SIGUSR2 recebido, eco em stderr desabilitado\n");
    bEcoTelaHabilitado = false;
  }
}


/*---------------------------------------------------------------------------------------------------------------------*
 * Funcoes de finalizacao                                                                                              *
 *---------------------------------------------------------------------------------------------------------------------*/
void EncerrarSimulacao()
{
  DestruirUniverso();
  if (ContactGroup != NULL)
    dJointGroupDestroy(ContactGroup);
  ContactGroup = NULL;
  LiberarMemoriaRobotnik();
  LiberarMemoriaAutomato();
  LiberarMemoriaPowell();
  if (bSimulacaoVisual)
    printf("\n");
  exit(0);
}


void TrataPedidoDestruicao(int iSinal) {
  static int iCount = 0;
  
  // Encerra o programa naturalmente quando o usuario der um kill
  if (iSinal == SIGTERM) {
    fprintf(stderr, "Sinal SIGTERM recebido, a aplicacao foi ABORTADA instantaneamente\n");
    exit(0);
  }
  else {
    iCount++;
    if (iCount > 1) {
      fprintf(stderr, "Sinal SIGINT recebido %d vezes, a aplicacao sera ABORTADA agora mesmo!\n", iCount);
      exit(0);
    }
    else {
      fprintf(stderr, "Sinal SIGINT recebido uma vez, AGUARDE o encerramento da aplicacao...\n");
      bAplicacaoCancelada = true;
    }
  }
}
