/*---------------------------------------------------------------------------------------------------------------------*
 * UNIVERSIDADE DO VALE DO RIO DOS SINOS - MESTRADO EM COMPUTACAO APLICADA                                             *
 * Sistemas Adaptativos Inteligentes - Prof. Fernando Santos Osorio                                                    *
 * Autor: Milton Roberto Heinen - Matr. 392201                                                                         *
 *---------------------------------------------------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------------------------------------------------*
 * Includes                                                                                                            *
 *---------------------------------------------------------------------------------------------------------------------*/
#include <ode/ode.h>
#include <drawstuff/drawstuff.h>
#include <ga/ga.h>
#define INSTANTIATE_REAL_GENOME
#include <ga/GARealGenome.h>
#include <signal.h>
#include <string>
using std::string;


/*---------------------------------------------------------------------------------------------------------------------*
 * Constantes                                                                                                          *
 *---------------------------------------------------------------------------------------------------------------------*/
// Constantes fisicas
const int MAX_CONTACTS = 4;
const int NUM_DIMENSOES = 3;
const double GRAVITY = -9.81;
const double ERP = 0.8;
const double CFM = 1.0e-4;
const double DENSIDADE = 0.5;
const double FORCA_MAX = 3500.0;
const double PASSO_FISICO = 0.05;
const double PASSO_AJUSTE = 0.2;

// Constantes da janela de visualizacao
const int TAM_TELA_X = 640;
const int TAM_TELA_Y = 480;

// Constantes dos sensores do robo
const double SAMPLE_GYRO_INTERVAL = 0.5;
const int MAX_SAMPLES_GYRO = 1024;
const int MAX_SAMPLES_BUMPERS = 10240;
const double PESO_GYRO = 10.0;

// Constantes gerais
const double NOVENTA_GRAUS_RAD = M_PI / 2.0;
const int MAX_LINHA = 5120;
const int NUM_FIRULAS = 4;
const double ATIVACAO_INICIAL = 0.0;


/*---------------------------------------------------------------------------------------------------------------------*
 * Variaveis globais                                                                                                   *
 *---------------------------------------------------------------------------------------------------------------------*/
// Mundo fisico
dWorldID World = NULL;
dBodyID *bodyMembros = NULL;
dBodyID *bodyJuntas = NULL;
dBodyID bodyCorpo;

// Espaca colisivo
dSpaceID Space = NULL;
dJointGroupID ContactGroup;
dGeomID *geomMembros = NULL;
dGeomID geomCorpo;
dGeomID geomSolo;

// Juntas
dJointID *hingeMembros = NULL;
dJointID *motorMembros = NULL;
dJointID *fixedMembros = NULL;

// Cabeca e rabo
dBodyID bodyFirulas[NUM_FIRULAS];
dGeomID geomFirulas[NUM_FIRULAS];
dJointID fixedFirulas[NUM_FIRULAS];

// Variaveis de drawstuff
dsFunctions myDrawFunction;
float CAMERA_XYZ[2][NUM_DIMENSOES] = {{  4.0f, 8.5f, 3.0f}, {  13.0f, 0.0f, 2.0f}};
float CAMERA_HPR[2][NUM_DIMENSOES] = {{-90.0f, 0.0f, 0.0f}, {-180.0f, 0.0f, 0.0f}};
char PATH_TEXTURES[] = "./textures";

// Variaveis do robotnik
string sNomeArquivoBase = "";
int iNumPernasRobotnik = 0;
int iNumPartesPorPerna = 0;
int iTotalPartesPernas = 0;
double **ppdTamPartesPernas = NULL;
double **ppdPosPartesPernas = NULL;
double **ppdPosJuntasPernas = NULL;
double **ppdRestricoesJunta = NULL;
double vdTamCorpoRobot[NUM_DIMENSOES];
double vdPosCorpoRobot[NUM_DIMENSOES];
bool bAngulosPatasCalculados = false;

// Cabeca e rabo
double vvdTamFirulas[NUM_FIRULAS][NUM_DIMENSOES] = {{0.6, 0.75, 1.0}, {1.65, 0.75, 0.6}, {0.5, 0.5, 0.5}, {0.5, 0.5, 1.5}};
double vvdPosFirulas[NUM_FIRULAS][NUM_DIMENSOES] = {{0.0, 0.0, 0.0}, {0.0, 0.0, 0.0}, {0.0, 0.0, 0.0}, {0.0, 0.0, 0.0}};

// Variaveis sensoriais
double vdGyroSamplesX[MAX_SAMPLES_GYRO];
double vdGyroSamplesY[MAX_SAMPLES_GYRO];
double vdGyroSamplesZ[MAX_SAMPLES_GYRO];
bool **pbBumperSamples = NULL;
bool *pbContatoComSolo = NULL;
int iCountGiroSamples = 0;
int iCountBumperSamples = 0;

// Variaveis da simulacao
bool bEcoTelaHabilitado = false;
bool bAplicacaoCancelada = false;
bool bSimulacaoVisual = true;
double dTimeSimulation = 0.0;
long lRandomSeed = 0;
bool bCaminhando = false;

// Variaveis geneticas
long lGeracaoAtual = 0;
long lEpochOfBest = 0;

// Variaveis da RNA
int iNumeroEntradas = 0;
int iNumeroSaidas = 0;
int iNumeroOcultos = 0;
int iNumeroPesos = 0;
double dInitPesos = 0.0;
double dFaixaPesos = 0.0;
double *pdEntradas = NULL;
double *pdSaidas = NULL;
double *pdOcultos = NULL;
double *pdContexto = NULL;
double **ppdPesosHidden = NULL;
double **ppdPesosSaida = NULL;


/*---------------------------------------------------------------------------------------------------------------------*
 * Prototipos                                                                                                          *
 *---------------------------------------------------------------------------------------------------------------------*/
// Funcoes estaticas ODE
static void NearCallback(void *xxx, dGeomID o1, dGeomID o2);
static void DrawstuffStart();
static void DrawstuffCommand(int cmd);
static void DrawstuffSimLoop(int pause);

// Funcoes do Robotnik
void CarregarRobotnik(const string sNomeArquivo);
void LiberarMemoriaRobotnik();

// Mundo fisico e colisivo
void CriaUniverso();
void CriaTodosObjetos();
void CriaJuntas();
void DestruirUniverso();

// Funcoes de controle da simulacao
void SetaPosOriginal();
double PhysicalSimLoop(const double dTimeSimulation);
void PararTodasAsJuntas();
inline void AjustarVelocidades();
inline double AjustarRestricoes(double dAnguloDesej, int iJunta);

// Funcoes sensoriais
double GetAlturaMaxBumpers();
double GetAlturaMinBumpers();
inline bool AlturaPataOk(const int iNumeroPata);
double CalcularInstabilidade();
double CalcularFatorBumpers();
double CalcularDeslocamento();

// Funcoes da RNA
void AlocarMemoriaRNA();
void CarregarPesosFromFile(const string sNomeArquivo);
void CarregarPesosFromGenoma(GAGenome &g);
inline void Ativar(const double *pdEntrada, double *pdSaida);
void SalvarPesos(const string sNomeArquivo);
void LiberarMemoriaRNA();

// Funcoes geneticas
void CarregarParametros(const string sArqParams);
void AlgoritmoGenetico(const string sArqParams);
static void Initializer(GAGenome& g);
static float Objective(GAGenome & g);

// Funcoes Utilitarias
string GetParamInFile(const string sNameFile, const string sParam);
inline double DegToRad(const double dValDeg);
inline double RadToDeg(const double dValRad);
inline double Quadrado(const double dValor);
inline double Normalizar(const double dValor, const double dMin, const double dMax);
inline double Anormaliza(const double dValor, const double dMin, const double dMax);
inline double Thereshold(const double dValor, const double dLimMin, const double dLimMax);
void HabilitaEcoTela(int iSinal);

// Funcoes de finalizacao
void EncerrarSimulacao();
void TrataPedidoDestruicao(int iSinal);


/*---------------------------------------------------------------------------------------------------------------------*
 * Funcao main                                                                                                         *
 *---------------------------------------------------------------------------------------------------------------------*/
int main(int argc, char *argv[])
{
  // Configura a manipulacao dos sinais de cancelamento
  struct sigaction saCancel;
  memset(&saCancel, 0, sizeof(saCancel));
  saCancel.sa_handler = &TrataPedidoDestruicao;
  sigaction(SIGINT, &saCancel, NULL);  
  sigaction(SIGTERM, &saCancel, NULL);
  
  // Configura a manipulacao dos sinais do usuario (eco)
  struct sigaction saSigusr;
  memset(&saSigusr, 0, sizeof(saSigusr));
  saSigusr.sa_handler = &HabilitaEcoTela;
  sigaction(SIGUSR1, &saSigusr, NULL);
  sigaction(SIGUSR2, &saSigusr, NULL);
  
  // Le os parametros da linha de comando
  if (argc > 1 && argv[1] && strncmp(argv[1], "-", 1))
    sNomeArquivoBase = argv[1];
  else {
    printf("Uso: %s <nome_arquivo> [-l]\n", argv[0]);
    return 1;
  }
  if (argc > 2 && argv[2] && !strcmp(argv[2], "-l"))
    bSimulacaoVisual = false;

  // Cria tudo
  CarregarRobotnik(sNomeArquivoBase + ".rbt");
  CriaUniverso();
  CriaTodosObjetos();
  CriaJuntas();

  if (bSimulacaoVisual) {
    // Visulizacao da simulacao
    CarregarPesosFromFile(sNomeArquivoBase + ".wts");
    dsSimulationLoop(argc, argv, TAM_TELA_X, TAM_TELA_Y, &myDrawFunction);
  }
  else {
    // Aprendizado genetico
    CarregarParametros(sNomeArquivoBase + ".par");
    AlocarMemoriaRNA();
    AlgoritmoGenetico(sNomeArquivoBase + ".par");
  }
  EncerrarSimulacao();
  return 0;
}


/*---------------------------------------------------------------------------------------------------------------------*
 * Funcoes estaticas ODE                                                                                               *
 *---------------------------------------------------------------------------------------------------------------------*/
static void NearCallback(void *xxx, dGeomID o1, dGeomID o2)
{
  static dContact contact[MAX_CONTACTS];
  register int i, iNumContacts = 0;
  static dBodyID b1, b2;
  static dJointID c;

  // Verifica se estao conectados
  b1 = dGeomGetBody(o1);
  b2 = dGeomGetBody(o2);
  if (b1 && b2 && dAreConnected(b1, b2))
    return;

  // Busca o numero de pontos de contato
  if (!(iNumContacts = dCollide(o1, o2, MAX_CONTACTS, &contact[0].geom, sizeof(dContact))))
    return;

  // Verifica se as patas entraram em contato com o solo neste turno (bumpers)
  if (o1 == geomSolo || o2 == geomSolo) {
    for (i = 0; i < iNumPernasRobotnik; i++) {
      if (!pbContatoComSolo[i] && (o1 == geomMembros[i * iNumPartesPorPerna] || o2  == geomMembros[i * iNumPartesPorPerna])) {
        pbContatoComSolo[i] = true;
        break;
      }
    }
  }

  for (i = 0; i < iNumContacts; i++) {
    // Configura a superficie de contato
    contact[i].surface.mode = dContactBounce | dContactSoftCFM;
    contact[i].surface.mu = dInfinity;
    contact[i].surface.mu2 = 0.0;
    contact[i].surface.bounce = 0.001;
    contact[i].surface.bounce_vel = 0.1;
    contact[i].surface.soft_cfm = 0.0001;

    // Cria as juntas do tipo contato para os pontos de colisao
    c = dJointCreateContact(World, ContactGroup, contact + i);
    dJointAttach(c, b1, b2);
  }
}


static void DrawstuffStart()
{
  // Inicializa a janela visual - seta a posicao da camera
  dsSetViewpoint(CAMERA_XYZ[0], CAMERA_HPR[0]);
}


static void DrawstuffCommand(int cmd)
{
  int i;

  // Interpretador de comandos do teclado
  if (cmd == 'x' || cmd == 'X') {
    PararTodasAsJuntas();
    EncerrarSimulacao();
  }
  else if (cmd == 's' || cmd == 'S')
    bCaminhando = true;
  else if (cmd == 'r' || cmd == 'R') {
    PararTodasAsJuntas();
    SetaPosOriginal();
    for (i = 0; i < iNumeroOcultos; i++)
      pdContexto[i] = ATIVACAO_INICIAL;
  }
  else if (cmd == 'v' || cmd == 'V')
    dsSetViewpoint(CAMERA_XYZ[0], CAMERA_HPR[0]);
  else if (cmd == 'f' || cmd == 'F')
    dsSetViewpoint(CAMERA_XYZ[1], CAMERA_HPR[1]);
}


static void DrawstuffSimLoop(int pause)
{
  register int i;
  struct timespec ts;
  static double dRelogioSimulacao = 0.0;
  static double dAtualizacaoAnterior = -PASSO_AJUSTE;
  static clock_t clAnterior = clock();
  clock_t clAtual = clock();
  dVector3 sizeBox;

  // Da uma pausa para que o tempo de exibicao se equipare ao tempo real de simulacao
  ts.tv_sec = 0;
  ts.tv_nsec = (long) ((PASSO_FISICO - (double) (clAtual - clAnterior) / (double) CLOCKS_PER_SEC) * 1000000000.0);
  nanosleep(&ts, NULL);
  clAnterior = clock();

  // Laco da simulacao fisico
  if (!pause && bCaminhando) {
    if (dRelogioSimulacao - dAtualizacaoAnterior >= PASSO_AJUSTE) {
      AjustarVelocidades();
      dAtualizacaoAnterior = dRelogioSimulacao;
    }
    dSpaceCollide(Space, 0, NearCallback);
    dWorldStep(World, PASSO_FISICO);
    dJointGroupEmpty(ContactGroup);
    dRelogioSimulacao += PASSO_FISICO;
  }

  // Desenha o robo
  dsSetTexture(DS_WOOD);
  dsSetColor(0.25, 1.0, 0.5);
  // Mebros
  for (i = 0; i < iTotalPartesPernas; i++) {
    dGeomBoxGetLengths(geomMembros[i], sizeBox);
    dsDrawBoxD(dBodyGetPosition(bodyMembros[i]), dBodyGetRotation(bodyMembros[i]), sizeBox);
    dsDrawCylinderD(dBodyGetPosition(bodyJuntas[i]), dBodyGetRotation(bodyJuntas[i]), ppdTamPartesPernas[1][1],
        ppdTamPartesPernas[1][0] / 2.0);
  }
  // Corpo
  dGeomBoxGetLengths(geomCorpo, sizeBox);
  dsDrawBoxD(dBodyGetPosition(bodyCorpo), dBodyGetRotation(bodyCorpo), sizeBox);
  // Cabeca e rabo
  for (i = 0; i < NUM_FIRULAS; i++) {
    dGeomBoxGetLengths(geomFirulas[i], sizeBox);
    dsDrawBoxD(dBodyGetPosition(bodyFirulas[i]), dBodyGetRotation(bodyFirulas[i]), sizeBox);
  }

  // Exibe o tempo de simulacao na tela
  printf("\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b");
  printf("Tempo: %02d:%02d   -   Deslocamento:%6.2f", (int) dRelogioSimulacao % 3600 / 60, (int) dRelogioSimulacao % 60,
      CalcularDeslocamento());
  fflush(stdout);
}


/*---------------------------------------------------------------------------------------------------------------------*
 * Funcoes do Robotnik                                                                                                 *
 *---------------------------------------------------------------------------------------------------------------------*/
void CarregarRobotnik(const string sNomeArquivo)
{
  char vcLinha[MAX_LINHA + 1];
  char *szPalavra = NULL;
  FILE *fp = NULL;
  int iCount, i, j, k;

  // Le o arquivo com as especificacoes do robotnik
  if ((fp = fopen(sNomeArquivo.c_str(), "r")) == NULL) {
    printf("Arquivo do robotnik \'%s\' inexistente\n", sNomeArquivo.c_str());
    exit(1);
  }

  // Le a primeira linha e busca o numero de pernas, o numero de partes por perna e o sincronismo
  fgets(vcLinha, MAX_LINHA, fp);
  if ((szPalavra = strtok(vcLinha, " ")) == NULL) {
    printf("Arquivo do robotnik \'%s\' incorreto\n", sNomeArquivo.c_str());
    exit(1);
  }
  iNumPernasRobotnik = atoi(szPalavra);
  if ((szPalavra = strtok('\0', " ")) == NULL) {
    printf("Arquivo do robotnik \'%s\' incorreto\n", sNomeArquivo.c_str());
    exit(1);
  }
  iNumPartesPorPerna = atoi(szPalavra);
  iTotalPartesPernas = iNumPartesPorPerna * iNumPernasRobotnik;
  szPalavra = strtok('\0', " ");
  if (szPalavra && (szPalavra = strtok('\0', " ")) != NULL && strlen(szPalavra) && (szPalavra[0] == 'S' || szPalavra[0] == 's'))
    bAngulosPatasCalculados = true;

  // Aloca memoria para o robotnik
  pbContatoComSolo = new bool[iNumPernasRobotnik];
  ppdTamPartesPernas = new double*[iTotalPartesPernas];
  ppdPosPartesPernas = new double*[iTotalPartesPernas];
  ppdPosJuntasPernas = new double*[iTotalPartesPernas];
  ppdRestricoesJunta = new double*[iTotalPartesPernas];
  for (i = 0; i < iTotalPartesPernas; i++) {
    ppdTamPartesPernas[i] = new double[NUM_DIMENSOES];
    ppdPosPartesPernas[i] = new double[NUM_DIMENSOES];
    ppdPosJuntasPernas[i] = new double[NUM_DIMENSOES];
    ppdRestricoesJunta[i] = new double[2];
  }

  // Inicializa algumas variaveis
  for (i = 0; i < iNumPernasRobotnik; i++)
    pbContatoComSolo[i] = false;

  // Aloca memoria para os bumpers
  pbBumperSamples = new bool*[MAX_SAMPLES_BUMPERS];
  for (i = 0; i < MAX_SAMPLES_BUMPERS; i++)
    pbBumperSamples[i] = new bool[iNumPernasRobotnik];

  // Le a segunda linha e pega o tamanho do corpo do robotnik
  fgets(vcLinha, MAX_LINHA, fp);
  szPalavra = strtok(vcLinha, " ");
  for (i = 0; i < NUM_DIMENSOES && szPalavra != NULL; i++) {
    vdTamCorpoRobot[i] = atof(szPalavra);
    szPalavra = strtok('\0', " ");
  }

  // Le a terceira linha e pega a posicao do corpo do robotnik
  fgets(vcLinha, MAX_LINHA, fp);
  szPalavra = strtok(vcLinha, " ");
  for (i = 0; i < NUM_DIMENSOES && szPalavra != NULL; i++) {
    vdPosCorpoRobot[i] = atof(szPalavra);
    szPalavra = strtok('\0', " ");
  }

  // Le as linhas subsequentes para pegar o tamanho das partes das pernas do robotnik
  iCount = 0;
  for (i = 0; i < iNumPernasRobotnik && !feof(fp); i++) {
    fgets(vcLinha, MAX_LINHA, fp);
    szPalavra = strtok(vcLinha, " ");
    for (j = 0; j < iNumPartesPorPerna && szPalavra != NULL; j++) {
      for (k = 0; k < NUM_DIMENSOES && szPalavra != NULL; k++) {
        ppdTamPartesPernas[iCount][k] = atof(szPalavra);
        szPalavra = strtok('\0', " ");
      }
      iCount++;
    }
  }

  // Le as ultimas linhas para pegar as restricoes das juntas do robotnik
  iCount = 0;
  for (i = 0; i < iNumPernasRobotnik && !feof(fp); i++) {
    fgets(vcLinha, MAX_LINHA, fp);
    szPalavra = strtok(vcLinha, " ");
    for (j = 0; j < iNumPartesPorPerna && szPalavra != NULL; j++) {
      for (k = 0; k < 2 && szPalavra != NULL; k++) {
        ppdRestricoesJunta[iCount][k] = atof(szPalavra);
        szPalavra = strtok('\0', " ");
      }
      iCount++;
    }
  }
  fclose(fp);

  // Preenche as posicoes das partes do robotnik
  iCount = 0;
  for (i = 0; i < iNumPernasRobotnik; i++) {
    for (j = 0; j < iNumPartesPorPerna; j++) {
      // Coordenada x
      if ((iNumPernasRobotnik == 4 && i % 2) || (iNumPernasRobotnik == 6 && (i == 0 || i == 3)))
        ppdPosPartesPernas[iCount][0] = vdTamCorpoRobot[0] / 2.0 - ppdTamPartesPernas[i * iNumPartesPorPerna +
            iNumPartesPorPerna - 1][0] / 2.0;
      else if ((iNumPernasRobotnik == 4) || (iNumPernasRobotnik == 6 && (i == 2 || i == 5)))
        ppdPosPartesPernas[iCount][0] = -vdTamCorpoRobot[0] / 2.0 + ppdTamPartesPernas[i * iNumPartesPorPerna +
            iNumPartesPorPerna - 1][0] / 2.0;
      else if (iNumPernasRobotnik == 6 || iNumPernasRobotnik == 2)
        ppdPosPartesPernas[iCount][0] = 0;

      // Coordenada y
      if (i < iNumPernasRobotnik / 2)
        ppdPosPartesPernas[iCount][1] = vdTamCorpoRobot[1] / 2.0 - ppdTamPartesPernas[i * iNumPartesPorPerna +
            iNumPartesPorPerna - 1][1] / 2.0;
      else
        ppdPosPartesPernas[iCount][1] = -vdTamCorpoRobot[1] / 2.0 + ppdTamPartesPernas[i * iNumPartesPorPerna +
            iNumPartesPorPerna - 1][1] / 2.0;
      // Coordenada z
      ppdPosPartesPernas[iCount][2] = 0.0;
      for (k = 0; k < j; k++)
        ppdPosPartesPernas[iCount][2] += ppdTamPartesPernas[i * iNumPartesPorPerna + k][2];
      ppdPosJuntasPernas[iCount][2] = ppdPosPartesPernas[iCount][2] + ppdTamPartesPernas[i * iNumPartesPorPerna + j][2];
      ppdPosPartesPernas[iCount][2] += ppdTamPartesPernas[i * iNumPartesPorPerna + j][2] / 2.0;
      // Juntas - x e y
      for (k = 0; k < NUM_DIMENSOES - 1; k++)
        ppdPosJuntasPernas[iCount][k] = ppdPosPartesPernas[iCount][k];
      iCount++;
    }
  }

  // Ajusta os tamanhos e as posicoes da cabeca e do rabo
  vvdTamFirulas[3][2] = vdTamCorpoRobot[2];
  vvdPosFirulas[0][0] = vdTamCorpoRobot[0] / 2.0 - vvdTamFirulas[0][0] / 2.0;
  vvdPosFirulas[1][0] = vdTamCorpoRobot[0] / 2.0 + vvdTamFirulas[1][0] / 2.0 - vvdTamFirulas[0][0];
  vvdPosFirulas[2][0] = -vdTamCorpoRobot[0] / 2.0 - vvdTamFirulas[2][0] / 2.0;
  vvdPosFirulas[3][0] = -vdTamCorpoRobot[0] / 2.0 - vvdTamFirulas[2][0] - vvdTamFirulas[3][0] / 2.0;
  vvdPosFirulas[0][2] = vdPosCorpoRobot[2] + vdTamCorpoRobot[2] / 2.0 +  vvdTamFirulas[0][2] / 2.0;
  vvdPosFirulas[1][2] = vvdPosFirulas[0][2] + vvdTamFirulas[0][2] / 2.0 + vvdTamFirulas[1][2] / 2.0;
  vvdPosFirulas[2][2] = vdPosCorpoRobot[2] + vdTamCorpoRobot[2] / 2.0 - vvdTamFirulas[2][2] / 2.0;
  vvdPosFirulas[3][2] = vdPosCorpoRobot[2];
}


void LiberarMemoriaRobotnik()
{
  int i;

  // Libera a memoria do robot e do automato
  if (ppdTamPartesPernas != NULL) {
    for (i = 0; i < iTotalPartesPernas; i++) {
      if (ppdTamPartesPernas[i] != NULL)
        delete [] ppdTamPartesPernas[i];
    }
    delete [] ppdTamPartesPernas;
    ppdTamPartesPernas = NULL;
  }
  if (ppdPosPartesPernas != NULL) {
    for (i = 0; i < iTotalPartesPernas; i++) {
      if (ppdPosPartesPernas[i] != NULL)
        delete [] ppdPosPartesPernas[i];
    }
    delete [] ppdPosPartesPernas;
    ppdPosPartesPernas = NULL;
  }
  if (ppdPosJuntasPernas != NULL) {
    for (i = 0; i < iTotalPartesPernas; i++) {
      if (ppdPosJuntasPernas[i] != NULL)
        delete [] ppdPosJuntasPernas[i];
    }
    delete [] ppdPosJuntasPernas;
    ppdPosJuntasPernas = NULL;
  }
  if (ppdRestricoesJunta != NULL) {
    for (i = 0; i < iTotalPartesPernas; i++) {
      if (ppdRestricoesJunta[i] != NULL)
        delete [] ppdRestricoesJunta[i];
    }
    delete [] ppdRestricoesJunta;
    ppdRestricoesJunta = NULL;
  }
  if (pbContatoComSolo != NULL)
    delete [] pbContatoComSolo;
  pbContatoComSolo = NULL;
  if (pbBumperSamples != NULL) {
    for (i = 0; i < MAX_SAMPLES_BUMPERS; i++)
      if (pbBumperSamples[i] != NULL)
        delete [] pbBumperSamples[i];
    delete [] pbBumperSamples;
  }
}


/*---------------------------------------------------------------------------------------------------------------------*
 * Mundo fisico e colisivo                                                                                             *
 *---------------------------------------------------------------------------------------------------------------------*/
void CriaUniverso()
{
  if (bSimulacaoVisual) {
    // Seta os parametros da funcao de fisualizacao
    myDrawFunction.version = DS_VERSION;
    myDrawFunction.start = &DrawstuffStart;
    myDrawFunction.step = &DrawstuffSimLoop;
    myDrawFunction.command = &DrawstuffCommand;
    myDrawFunction.path_to_textures = PATH_TEXTURES;
  }

  // Criacao e inicializacao do mundo fisico
  World = dWorldCreate();
  dWorldSetGravity(World, 0.0, 0.0, GRAVITY);
  dWorldSetERP(World, ERP);
  dWorldSetCFM(World, CFM);

  // Criacao do espaco de colisao, do grupo de contatos e do solo
  Space = dSimpleSpaceCreate(0);
  ContactGroup = dJointGroupCreate(0);
  geomSolo = dCreatePlane(Space, 0.0, 0.0, 1.0, 0.0);
}


void CriaTodosObjetos()
{
  int i;
  dMass massObjects;
  dMatrix3 rotacaoMatrix;

  // ********** Cria os membros ********** //
  bodyMembros = new dBodyID[iTotalPartesPernas];
  geomMembros = new dGeomID[iTotalPartesPernas];
  for (i = 0; i < iTotalPartesPernas; i++) {
    // Parte fisica
    bodyMembros[i] = dBodyCreate(World);
    dMassSetBox(&massObjects, DENSIDADE, ppdTamPartesPernas[i][0], ppdTamPartesPernas[i][1], ppdTamPartesPernas[i][2]);
    dBodySetMass(bodyMembros[i], &massObjects);
    dBodySetPosition(bodyMembros[i], ppdPosPartesPernas[i][0], ppdPosPartesPernas[i][1], ppdPosPartesPernas[i][2]);
    // Parte colisiva
    geomMembros[i] = dCreateBox(Space, ppdTamPartesPernas[i][0], ppdTamPartesPernas[i][1], ppdTamPartesPernas[i][2]);
    dGeomSetPosition(geomMembros[i], ppdPosPartesPernas[i][0], ppdPosPartesPernas[i][1], ppdPosPartesPernas[i][2]);
    // Conecta a parte fisica com a colisiva
    dGeomSetBody(geomMembros[i], bodyMembros[i]);
  }

  // ********** Cria o corpo ********** //
  // Parte fisica
  bodyCorpo = dBodyCreate(World);
  dMassSetBox(&massObjects, DENSIDADE, vdTamCorpoRobot[0], vdTamCorpoRobot[1], vdTamCorpoRobot[2]);
  dBodySetMass(bodyCorpo, &massObjects);
  dBodySetPosition(bodyCorpo, vdPosCorpoRobot[0], vdPosCorpoRobot[1], vdPosCorpoRobot[2]);
  // Parte colisiva
  geomCorpo = dCreateBox(Space, vdTamCorpoRobot[0], vdTamCorpoRobot[1], vdTamCorpoRobot[2]);
  dGeomSetPosition(geomCorpo, vdPosCorpoRobot[0], vdPosCorpoRobot[1], vdPosCorpoRobot[2]);
  // Conecta a parte fisica com a colisiva
  dGeomSetBody(geomCorpo, bodyCorpo);

  // ********** Cria as juntas visuais ********** //
  if (bSimulacaoVisual) {
    dRFromAxisAndAngle(rotacaoMatrix, 1, 0, 0, NOVENTA_GRAUS_RAD);
    bodyJuntas = new dBodyID[iTotalPartesPernas];
    for (i = 0; i < iTotalPartesPernas; i++) {
      bodyJuntas[i] = dBodyCreate(World);
      dMassSetCylinder(&massObjects, DENSIDADE / 1000.0, 2, ppdTamPartesPernas[1][1], ppdTamPartesPernas[1][0] / 2.0);
      dBodySetMass(bodyJuntas[i], &massObjects);
      dBodySetPosition(bodyJuntas[i], ppdPosJuntasPernas[i][0], ppdPosJuntasPernas[i][1], ppdPosJuntasPernas[i][2]);
      dBodySetRotation(bodyJuntas[i], rotacaoMatrix);
    }

    // Cabeca e rabo
    for (i = 0; i < NUM_FIRULAS; i++) {
      bodyFirulas[i] = dBodyCreate(World);
      dMassSetBox(&massObjects, DENSIDADE / 1000.0, vvdTamFirulas[i][0], vvdTamFirulas[i][1], vvdTamFirulas[i][2]);
      dBodySetMass(bodyFirulas[i], &massObjects);
      dBodySetPosition(bodyFirulas[i], vvdPosFirulas[i][0], vvdPosFirulas[i][1], vvdPosFirulas[i][2]);
      geomFirulas[i] = dCreateBox(Space, vvdTamFirulas[i][0], vvdTamFirulas[i][1], vvdTamFirulas[i][2]);
      dGeomSetPosition(geomFirulas[i], vvdPosFirulas[i][0], vvdPosFirulas[i][1], vvdPosFirulas[i][2]);
      dGeomSetBody(geomFirulas[i], bodyFirulas[i]);
      fixedFirulas[i] = dJointCreateFixed(World, 0);
      dJointAttach(fixedFirulas[i], bodyFirulas[i], bodyCorpo);
      dJointSetFixed(fixedFirulas[i]);
    }
  }
}


void CriaJuntas()
{
  int i;

  // Cria as juntas e motores
  hingeMembros = new dJointID[iTotalPartesPernas];
  motorMembros = new dJointID[iTotalPartesPernas];
  fixedMembros = new dJointID[iTotalPartesPernas];

  for (i = 0; i < iTotalPartesPernas; i++) {
    // Cria as juntas
    hingeMembros[i] = dJointCreateHinge(World, 0);
    dJointAttach(hingeMembros[i], bodyMembros[i], (((i + 1) % iNumPartesPorPerna) ? bodyMembros[i + 1] : bodyCorpo));
    dJointSetHingeAxis(hingeMembros[i], 0.0, 1.0, 0.0);
    dJointSetHingeAnchor(hingeMembros[i], ppdPosJuntasPernas[i][0], ppdPosJuntasPernas[i][1], ppdPosJuntasPernas[i][2]);
    dJointSetHingeParam(hingeMembros[i], dParamBounce, 0.0);
    dJointSetHingeParam(hingeMembros[i], dParamFMax, 0.0);
    // O ponto zero tem que estar dentro da faixa de restricoes (senao o modelo explode)
    if (ppdRestricoesJunta[i][0] > 0.0)
      dJointSetHingeParam(hingeMembros[i], dParamLoStop, 0.0);
    else
      dJointSetHingeParam(hingeMembros[i], dParamLoStop, ppdRestricoesJunta[i][0]);
    if (ppdRestricoesJunta[i][1] < 0.0)
      dJointSetHingeParam(hingeMembros[i], dParamHiStop, 0.0);
    else
      dJointSetHingeParam(hingeMembros[i], dParamHiStop, ppdRestricoesJunta[i][1]);

    // Cria os motores das juntas
    motorMembros[i] = dJointCreateAMotor(World, 0);
    dJointSetAMotorMode(motorMembros[i], dAMotorUser);
    dJointAttach(motorMembros[i], bodyMembros[i], (((i + 1) % iNumPartesPorPerna) ? bodyMembros[i + 1] : bodyCorpo));
    dJointSetAMotorNumAxes(motorMembros[i], 1);
    dJointSetAMotorAxis(motorMembros[i], 0, 0, 0.0, 1.0, 0.0);
    dJointSetAMotorParam(motorMembros[i], dParamFMax, FORCA_MAX);
    dJointSetAMotorParam(motorMembros[i], dParamVel, 0.0);

    // Cria cilindros fisicos nas juntas com densidade pequena
    if (bSimulacaoVisual) {
      fixedMembros[i] = dJointCreateFixed(World, 0);
      dJointAttach(fixedMembros[i], bodyJuntas[i], bodyMembros[i]);
      dJointSetFixed(fixedMembros[i]);
    }
  }
}


void DestruirUniverso()
{
  // Apaga os objetos ODE alocados dinamicamente
  if (bodyMembros != NULL)
    delete [] bodyMembros;
  bodyMembros = NULL;
  if (geomMembros != NULL)
    delete [] geomMembros;
  geomMembros = NULL;
  if (hingeMembros != NULL)
    delete [] hingeMembros;
  hingeMembros = NULL;
  if (motorMembros != NULL)
    delete [] motorMembros;
  motorMembros = NULL;
  if (fixedMembros != NULL)
    delete [] fixedMembros;
  fixedMembros = NULL;
  if (bodyJuntas != NULL)
    delete [] bodyJuntas;
  bodyJuntas = NULL;
  if (Space != NULL)
    dSpaceDestroy(Space);
  Space = NULL;
  if (World != NULL)
    dWorldDestroy(World);
  World = NULL;
  if (ContactGroup != NULL)
    dJointGroupDestroy(ContactGroup);
  ContactGroup = NULL;
}


/*---------------------------------------------------------------------------------------------------------------------*
 * Funcoes de controle da simulacao                                                                                    *
 *---------------------------------------------------------------------------------------------------------------------*/
void SetaPosOriginal()
{
  int i;
  dQuaternion quatObj;
  dMatrix3 rotacaoMatrix;

  // Restaura a posicao original e atitude dos objetos fisicos
  dQSetIdentity(quatObj);
  for (i = 0; i < iTotalPartesPernas; i++) {
    dBodySetPosition(bodyMembros[i], ppdPosPartesPernas[i][0], ppdPosPartesPernas[i][1], ppdPosPartesPernas[i][2]);
    dBodySetQuaternion(bodyMembros[i], quatObj);
    dGeomSetPosition(geomMembros[i], ppdPosPartesPernas[i][0], ppdPosPartesPernas[i][1], ppdPosPartesPernas[i][2]);
    dGeomSetQuaternion(geomMembros[i], quatObj);
  }
  dBodySetPosition(bodyCorpo, vdPosCorpoRobot[0], vdPosCorpoRobot[1], vdPosCorpoRobot[2]);
  dBodySetQuaternion(bodyCorpo, quatObj);
  dGeomSetPosition(geomCorpo, vdPosCorpoRobot[0], vdPosCorpoRobot[1], vdPosCorpoRobot[2]);
  dGeomSetQuaternion(geomCorpo, quatObj);

  if (bSimulacaoVisual) {
    dRFromAxisAndAngle(rotacaoMatrix, 1, 0, 0, NOVENTA_GRAUS_RAD);
    for (i = 0; i < iTotalPartesPernas; i++) {
      dBodySetPosition(bodyJuntas[i], ppdPosJuntasPernas[i][0], ppdPosJuntasPernas[i][1], ppdPosJuntasPernas[i][2]);
      dBodySetRotation(bodyJuntas[i], rotacaoMatrix);
    }
    for (i = 0; i < NUM_FIRULAS; i++) {
      dBodySetPosition(bodyFirulas[i], vvdPosFirulas[i][0], vvdPosFirulas[i][1], vvdPosFirulas[i][2]);
      dBodySetQuaternion(bodyFirulas[i], quatObj);
      dGeomSetPosition(geomFirulas[i], vvdPosFirulas[i][0], vvdPosFirulas[i][1], vvdPosFirulas[i][2]);
      dGeomSetQuaternion(geomFirulas[i], quatObj);
    }
  }
}


double PhysicalSimLoop(const double dTimeSimulation)
{
  const dReal *corpoPosAtual = dBodyGetPosition(bodyCorpo);
  double dPosAntX = corpoPosAtual[0], dPosAntY = corpoPosAtual[1], dPosAntZ = corpoPosAtual[2];
  double dRelogioSimulacao = 0.0, dTimeAntGiro = 0.0, dAtualizacaoAnterior = -PASSO_AJUSTE;
  register bool bRobotnikCaindo = true;
  register int i;


  // Inicializacoes dos sensores
  iCountBumperSamples = iCountGiroSamples = 0;
  for (i = 0; i < iNumPernasRobotnik; i++)
    pbContatoComSolo[i] = false;
  bCaminhando = true;

  // Laco de simulacao fisico
  while (dRelogioSimulacao < dTimeSimulation) {
    if (dRelogioSimulacao - dAtualizacaoAnterior >= PASSO_AJUSTE) {
      AjustarVelocidades();
      dAtualizacaoAnterior = dRelogioSimulacao;
    }
    dSpaceCollide(Space, 0, NearCallback);
    dWorldStep(World, PASSO_FISICO);
    dJointGroupEmpty(ContactGroup);
    dRelogioSimulacao += PASSO_FISICO;

    // Leitura do giroscopio
    if (dRelogioSimulacao - dTimeAntGiro > SAMPLE_GYRO_INTERVAL && iCountGiroSamples < MAX_SAMPLES_GYRO) {
      corpoPosAtual = dBodyGetPosition(bodyCorpo);
      vdGyroSamplesX[iCountGiroSamples] = (corpoPosAtual[0] - dPosAntX);
      vdGyroSamplesY[iCountGiroSamples] = (corpoPosAtual[1] - dPosAntY);
      vdGyroSamplesZ[iCountGiroSamples] = (corpoPosAtual[2] - dPosAntZ);
      dPosAntX = corpoPosAtual[0];
      dPosAntY = corpoPosAtual[1];
      dPosAntZ = corpoPosAtual[2];
      dTimeAntGiro = dRelogioSimulacao;
      iCountGiroSamples++;
    }

    // Leitura dos bumpers
    bRobotnikCaindo = true;
    if (iCountBumperSamples < MAX_SAMPLES_BUMPERS) {
      for (i = 0; i < iNumPernasRobotnik; i++) {
        pbBumperSamples[iCountBumperSamples][i] = pbContatoComSolo[i] && AlturaPataOk(i);
        if (bRobotnikCaindo && pbBumperSamples[iCountBumperSamples][i])
          bRobotnikCaindo = false;
        pbContatoComSolo[i] = false;
      }
      iCountBumperSamples++;
      if (bRobotnikCaindo)
        break;
    }
  }
  PararTodasAsJuntas();
  return dRelogioSimulacao;
}


void PararTodasAsJuntas()
{
  int i;

  // Para todas as juntas do robo
  bCaminhando = false;
  for (i = 0; i < iTotalPartesPernas; i++)
    dJointSetAMotorParam(motorMembros[i], dParamVel, 0.0);
}


inline void AjustarVelocidades()
{
  register int i, j, k, iCount = 0;
  double dAngleDesej, dSomaAngulo, dVelocidade;

  // Prenche o vetor de entradas com os angulos atuais e ativa a RNA
  for (i = 0; i < iNumPernasRobotnik / 2; i++) {
    for (j = 1; j < iNumPartesPorPerna; j++)
      pdEntradas[iCount++] = Normalizar(dJointGetHingeAngle(hingeMembros[i * iNumPartesPorPerna + j]),
          ppdRestricoesJunta[i * iNumPartesPorPerna + j][0], ppdRestricoesJunta[i * iNumPartesPorPerna + j][1]);
  }
  Ativar(pdEntradas, pdSaidas);

  // Ajusta as velocidades das juntas, menos a da pata (mas calcula a velocidade ideal da pata)
  for (i = 0; i < iNumPernasRobotnik; i++) {
    k  = ((i >= iNumPernasRobotnik / 2) ? abs(i - 3) : i);
    dSomaAngulo = 0.0;
    for (j = 1; j < iNumPartesPorPerna; j++) {
      dAngleDesej = Anormaliza(pdSaidas[k * (iNumPartesPorPerna - 1) + j - 1], ppdRestricoesJunta[i * iNumPartesPorPerna + j][0],
          ppdRestricoesJunta[i * iNumPartesPorPerna + j][1]);
      dSomaAngulo += dAngleDesej;
      dVelocidade = dAngleDesej - dJointGetHingeAngle(hingeMembros[i * iNumPartesPorPerna + j]);
      dJointSetAMotorParam(motorMembros[i * iNumPartesPorPerna + j], dParamVel, dVelocidade);
    }
    // Ajusta as velocidades da pata
    dAngleDesej = AjustarRestricoes(-dSomaAngulo, i * iNumPartesPorPerna);
    dVelocidade = dAngleDesej - dJointGetHingeAngle(hingeMembros[i * iNumPartesPorPerna]);
    dJointSetAMotorParam(motorMembros[i * iNumPartesPorPerna], dParamVel, dVelocidade);
  }
}


inline double AjustarRestricoes(double dAnguloDesej, int iJunta)
{
  // Verifica se os angulos desejados violam as restricoes da junta - retorna o angulo ajustado
  if (dAnguloDesej < ppdRestricoesJunta[iJunta][0])
    return ppdRestricoesJunta[iJunta][0];
  else if (dAnguloDesej > ppdRestricoesJunta[iJunta][1])
    return ppdRestricoesJunta[iJunta][1];
  else
    return dAnguloDesej;
}


/*---------------------------------------------------------------------------------------------------------------------*
 * Funcoes sensoriais                                                                                                  *
 *---------------------------------------------------------------------------------------------------------------------*/
double GetAlturaMaxBumpers()
{
  const double dInclinacaoMax = DegToRad(7.5);
  double dAngulo, dAlturaMax;

  // Calcula a altura maxima para o pe ser considerado no chao
  dAngulo = atan(ppdTamPartesPernas[0][2] / ppdTamPartesPernas[0][0]);
  if (ppdTamPartesPernas[0][2] > ppdTamPartesPernas[0][0]) // Pes mais altos do que largos (cavalo)
    dAlturaMax = (ppdTamPartesPernas[0][2] / 2.0) / sin(dAngulo);
  else  // Pes mais largos do que altos (humanos)
    dAlturaMax = sin(dInclinacaoMax + dAngulo) * ((ppdTamPartesPernas[0][0] / 2.0) / cos(dAngulo));
  return dAlturaMax;
}


double GetAlturaMinBumpers()
{
  const double dInclinacaoMax = DegToRad(15.0);
  double dAngulo, dAlturaMin = 0.0;
  double dAlturaMax = GetAlturaMaxBumpers();

  // Calcula a altura minima para o pe ser considerado no chao
  dAngulo = atan(ppdTamPartesPernas[0][0] / ppdTamPartesPernas[0][2]);
  if (ppdTamPartesPernas[0][2] > ppdTamPartesPernas[0][0])
    dAlturaMin = sin(dInclinacaoMax + dAngulo) * dAlturaMax;
  return dAlturaMin;
}


inline bool AlturaPataOk(const int iNumeroPata)
{
  static double dAlturaMax = GetAlturaMaxBumpers();
  static double dAlturaMin = GetAlturaMinBumpers();
  double dAlturaAtual;

  // Conta o numero de pes que estao em contato com o chao
  dAlturaAtual = dBodyGetPosition(bodyMembros[iNumeroPata * iNumPartesPorPerna])[2];
  if (dAlturaAtual <= dAlturaMax && dAlturaAtual >= dAlturaMin)
    return true;
  else
    return false;
}


double CalcularInstabilidade()
{
  double dMediaX, dMediaY, dMediaZ, dVarianceX, dVarianceY, dVarianceZ;
  register int i;

  // Calcula a instabilidade do caminhar
  if (iCountGiroSamples < 2)
    return 1.0;

  // Calcula as medias
  dMediaX = dMediaY = dMediaZ = 0.0;
  for (i = 0; i < iCountGiroSamples; i++) {
    dMediaX += vdGyroSamplesX[i];
    dMediaY += vdGyroSamplesY[i];
    dMediaZ += vdGyroSamplesZ[i];
  }
  dMediaX /= (double) iCountGiroSamples;
  dMediaY /= (double) iCountGiroSamples;
  dMediaZ /= (double) iCountGiroSamples;

  // Calcula as variancias
  dVarianceX = dVarianceY = dVarianceZ = 0.0;
  for (i = 0; i < iCountGiroSamples; i++) {
    dVarianceX += Quadrado(vdGyroSamplesX[i] - dMediaX);
    dVarianceY += Quadrado(vdGyroSamplesY[i] - dMediaY);
    dVarianceZ += Quadrado(vdGyroSamplesZ[i] - dMediaZ);
  }

  // Retorna o desvio padrao das leituras do giroscopio
  return sqrt((dVarianceX + dVarianceY + dVarianceZ) / (iCountGiroSamples - 1.0));
}


double CalcularFatorBumpers()
{
  const int iMaxAmostras = (int) (dTimeSimulation / PASSO_FISICO);
  register int i, j, k, iCount;
  double dFactor = 0.0;

  if (iCountBumperSamples < iNumPernasRobotnik)
    return 1.0;

  // O fator varia de 1 (tudo no chao/ar) a 0 (trote ultra cadenciado)
  for (j = 0; j < iNumPernasRobotnik; j++) {
    iCount = 0;
    k = (j + iNumPernasRobotnik / 2) % iNumPernasRobotnik;
    for (i = 0; i < iCountBumperSamples; i++) {
      if (pbBumperSamples[i][j] && !pbBumperSamples[i][k])
        iCount++;
    }
    dFactor += Quadrado((double) iCount / (double) iCountBumperSamples - 0.5);
  }

  // Faz com que o fator fique entre 0 e 1 independente do nmero de pernas do robotnik
  dFactor = dFactor * 4.0 / (double) iNumPernasRobotnik;

  // Se a simulacao nao foi ate o fim, descontamos o fator
  if (iCountBumperSamples < iMaxAmostras)
    dFactor = dFactor * ((double) iCountBumperSamples / (double) iMaxAmostras) + ((double) iMaxAmostras - (double)
      iCountBumperSamples) / (double) iMaxAmostras;
  return dFactor;
}


double CalcularDeslocamento()
{
  double dPosIni = vdPosCorpoRobot[0];
  double dPosFim = dBodyGetPosition(bodyCorpo)[0];
  register int i;

  for (i = 0; i < iNumPernasRobotnik; i++) {
    dPosIni += ppdPosPartesPernas[i * iNumPartesPorPerna][0];
    dPosFim += dBodyGetPosition(bodyMembros[i * iNumPartesPorPerna])[0];
  }
  return (dPosFim - dPosIni) / ((double) iNumPernasRobotnik + 1.0);
}


/*---------------------------------------------------------------------------------------------------------------------*
 * Funcoes da RNA                                                                                                      *
 *---------------------------------------------------------------------------------------------------------------------*/
void AlocarMemoriaRNA()
{
  int i;

  LiberarMemoriaRNA();
  iNumeroEntradas = (iNumPartesPorPerna - 1) * (iNumPernasRobotnik / 2);
  iNumeroSaidas = (iNumPartesPorPerna - 1) * (iNumPernasRobotnik / 2);
  iNumeroPesos = iNumeroOcultos * (iNumeroEntradas + iNumeroOcultos + 1) + iNumeroSaidas * (iNumeroOcultos + 1);
  pdEntradas = new double[iNumeroEntradas];
  pdOcultos = new double[iNumeroOcultos];
  pdContexto = new double[iNumeroOcultos];
  pdSaidas = new double[iNumeroSaidas];
  ppdPesosHidden = new double*[iNumeroOcultos];
  for (i = 0; i < iNumeroOcultos; i++) {
    ppdPesosHidden[i] = new double[iNumeroEntradas + iNumeroOcultos + 1];
    pdContexto[i] = ATIVACAO_INICIAL;
  }
  ppdPesosSaida = new double*[iNumeroSaidas];
  for (i = 0; i < iNumeroSaidas; i++)
    ppdPesosSaida[i] = new double[iNumeroOcultos + 1];
}


void CarregarPesosFromFile(const string sNomeArquivo)
{
  int i, iNeuronio = 0;
  FILE *fp = NULL;
  char vcLinha[MAX_LINHA + 1];
  char *szPalavra = NULL;

  if ((fp = fopen(sNomeArquivo.c_str(), "r")) == NULL) {
    printf("Nao foi possivel abrir o arquivo de pesos\n");
    exit(1);
  }

  // Busca o numero de entradas, ocultos e saidas
  fgets(vcLinha, MAX_LINHA, fp);
  szPalavra = strtok(vcLinha, " ");
  iNumeroEntradas = atoi(szPalavra);
  szPalavra = strtok('\0', " ");
  iNumeroOcultos = atoi(szPalavra);
  szPalavra = strtok('\0', " ");
  iNumeroSaidas = atoi(szPalavra);
  AlocarMemoriaRNA();

  // Carrega os pesos da camada oculta - um neuronio por linha
  while (!feof(fp) && iNeuronio < iNumeroOcultos) {
    fgets(vcLinha, MAX_LINHA, fp);
    szPalavra = strtok(vcLinha, " ");
    pdContexto[iNeuronio] = ATIVACAO_INICIAL;
    for (i = 0; i < iNumeroEntradas + iNumeroOcultos + 1 && szPalavra != NULL; i++) {
      ppdPesosHidden[iNeuronio][i] = atof(szPalavra);
      szPalavra = strtok('\0', " ");
    }
    iNeuronio++;
  }

  // Carrega os pesos da camada de saida - um neuronio por linha
  iNeuronio = 0;
  while (!feof(fp) && iNeuronio < iNumeroSaidas) {
    fgets(vcLinha, MAX_LINHA, fp);
    szPalavra = strtok(vcLinha, " ");
    for (i = 0; i <= iNumeroOcultos && szPalavra != NULL; i++) {
      ppdPesosSaida[iNeuronio][i] = atof(szPalavra);
      szPalavra = strtok('\0', " ");
    }
    iNeuronio++;
  }
  fclose(fp);
  printf("Arquivo de pesos \'%s\' carregado com sucesso\n", sNomeArquivo.c_str());
}


void CarregarPesosFromGenoma(GAGenome &g) {
  GARealGenome &genome = (GARealGenome&) g;
  int i, j, iCount = 0;

  // Carrega o genoma para os pesos da RNA
  for (i = 0; i < iNumeroOcultos; i++) {
    pdContexto[i] = ATIVACAO_INICIAL;
    for (j = 0; j < iNumeroEntradas + iNumeroOcultos + 1; j++)
      ppdPesosHidden[i][j] = (double) genome.gene(iCount++);
  }
  for (i = 0; i < iNumeroSaidas; i++) {
    for (j = 0; j <= iNumeroOcultos; j++)
      ppdPesosSaida[i][j] = (double) genome.gene(iCount++);
  }
}


inline void Ativar(const double *pdEntrada, double *pdSaida)
{
  register int i, j;

  // Ativacao da camada oculta
  for (i = 0; i < iNumeroOcultos; i++) {
    pdOcultos[i] = 0.0;
    for (j = 0; j < iNumeroEntradas; j++)
      pdOcultos[i] += pdEntrada[j] * ppdPesosHidden[i][j];
    for (j = 0; j < iNumeroOcultos; j++)
      pdOcultos[i] += pdContexto[i] * ppdPesosHidden[i][iNumeroEntradas + j];
    pdOcultos[i] = tanh(pdOcultos[i] + ppdPesosHidden[i][iNumeroEntradas + iNumeroOcultos]);
  }
  // Ativacao da camada de saida
  for (i = 0; i < iNumeroSaidas; i++) {
    pdSaida[i] = 0.0;
    for (j = 0; j < iNumeroOcultos; j++)
      pdSaida[i] += pdOcultos[j] * ppdPesosSaida[i][j];
    pdSaida[i] = tanh(pdSaida[i] + ppdPesosSaida[i][iNumeroOcultos]);
  }
  // Atualiza as unidades de contexto
  for (i = 0; i < iNumeroOcultos; i++)
    pdContexto[i] = pdOcultos[i];
}


void SalvarPesos(const string sNomeArquivo)
{
  int i, j;
  FILE *fp = NULL;

  // Abre o arquivo e imprime o numero de entradas, ocultos, saidas e pesos
  if ((fp = fopen(sNomeArquivo.c_str(), "w")) == NULL) {
    printf("Nao foi possivel criar o arquivo de pesos \'%s\'\n", sNomeArquivo.c_str());
    exit(1);
  }
  fprintf(fp, "%d %d %d %d\n", iNumeroEntradas, iNumeroOcultos, iNumeroSaidas, iNumeroPesos);
  // Salva os pesos ocultos
  for (i = 0; i < iNumeroOcultos; i++) {
    for (j = 0; j < iNumeroEntradas + iNumeroOcultos + 1; j++)
      fprintf(fp, "%12.8f ", ppdPesosHidden[i][j]);
    fprintf(fp, "\n");
  }
  // Salva os pesos de saida e fecha o arquivo
  for (i = 0; i < iNumeroSaidas; i++) {
    for (j = 0; j <= iNumeroOcultos; j++)
      fprintf(fp, "%12.8f ", ppdPesosSaida[i][j]);
    fprintf(fp, "\n");
  }
  fclose(fp);
}


void LiberarMemoriaRNA()
{
  int i;

  if (ppdPesosHidden != NULL) {
    for (i = 0; i < iNumeroOcultos; i++)
      if (ppdPesosHidden[i] != NULL)
        delete [] ppdPesosHidden[i];
    delete [] ppdPesosHidden;
  }
  ppdPesosHidden = NULL;
  if (ppdPesosSaida != NULL) {
    for (i = 0; i < iNumeroSaidas; i++)
      if (ppdPesosSaida[i] != NULL)
        delete [] ppdPesosSaida[i];
    delete [] ppdPesosSaida;
  }
  ppdPesosSaida = NULL;
  if (pdEntradas != NULL)
    delete [] pdEntradas;
  pdEntradas = NULL;
  if (pdOcultos != NULL)
    delete [] pdOcultos;
  pdOcultos = NULL;
  if (pdContexto != NULL)
    delete [] pdContexto;
  pdContexto = NULL;
  if (pdSaidas != NULL)
    delete [] pdSaidas;
  pdSaidas = NULL;
}


/*---------------------------------------------------------------------------------------------------------------------*
 * Funcoes geneticas                                                                                                   *
 *---------------------------------------------------------------------------------------------------------------------*/
void CarregarParametros(const string sArqParams) {
  // Busca os parametros geneticos
  lRandomSeed = atol(GetParamInFile(sArqParams, "random_seed").c_str());
  dInitPesos = atof(GetParamInFile(sArqParams, "weights_init").c_str());
  dFaixaPesos = atof(GetParamInFile(sArqParams, "weights_interval").c_str());
  iNumeroOcultos = atoi(GetParamInFile(sArqParams, "hidden_neurons").c_str());
  dTimeSimulation = atof(GetParamInFile(sArqParams, "time_simulation").c_str());
}


void AlgoritmoGenetico(const string sArqParams)
{
  time_t tmAtual, tmInicial = time(NULL);
  long lTime, lTimeRestante;
  int i;
  char szMensagem[MAX_LINHA + 1];
  lEpochOfBest = lGeracaoAtual = 0;

  // Cria os alelos
  GARealAlleleSetArray alleles;
  for (i = 0; i < iNumeroPesos; i++)
    alleles.add(-dFaixaPesos, dFaixaPesos);

  // Cria o genoma
  GARealGenome genome(alleles, Objective);
  genome.initializer(Initializer);
  genome.comparator(GARealGenome::ElementComparator);
  genome.mutator(GARealGaussianMutator);
  genome.crossover(GARealGenome::OnePointCrossover);

  // Cria e inicializa o GA
  GASteadyStateGA ga(genome);
  ga.maximize();
  ga.parameters(sArqParams.c_str());
  ga.recordDiversity(gaFalse);
  ga.selectScores(GAStatistics::Maximum | GAStatistics::Mean | GAStatistics::Deviation);
  GASigmaTruncationScaling sigmaTruncation;
  ga.scaling(sigmaTruncation);
  GASRSSelector srsSelector;
  ga.selector(srsSelector);

  // Informacoes iniciais
  printf("Tamanho da Population..:%5d\n", ga.populationSize());
  printf("Numero de Generations..:%5d\n", ga.nGenerations());
  printf("Taxa de Mutation.......:%10.4f\n", ga.pMutation());
  printf("Taxa de Crossover......:%10.4f\n", ga.pCrossover());
  printf("Taxa de Replacement....:%10.4f\n", ga.pReplacement());
  printf("Random seed utilizado..:%5li\n", lRandomSeed);
  printf("Frequencia relatorios..:%5d\n", ga.scoreFrequency());
  printf("Numero de ocultos......:%5d\n", iNumeroOcultos);
  printf("Faixa dos pesos (-/+)..:%10.4f\n", dFaixaPesos);
  printf("Inicializacao pesos....:%10.4f\n", dInitPesos);
  printf("Tempo de caminhada.....:%10.4f\n", dTimeSimulation);

  // Realiza a evolucao natural
  ga.initialize(lRandomSeed);
  while (!ga.done() && (ga.statistics().current(GAStatistics::Deviation) > 1.0e-4 || lGeracaoAtual < ga.nGenerations() / 5) 
      && !bAplicacaoCancelada && lGeracaoAtual - lEpochOfBest < ga.nGenerations() / 6) {
    // Passo evolutivo
    ga.step();
    lGeracaoAtual = ga.generation();

    // Exibe relatorio da evolucao
    if (!(lGeracaoAtual % ga.scoreFrequency())) {
      tmAtual = time(NULL);
      lTime = (long) (tmAtual - tmInicial);
      lTimeRestante = (long) ((double) ((tmAtual - tmInicial) / (lGeracaoAtual + 1.0)) * (ga.nGenerations() - lGeracaoAtual));
      sprintf(szMensagem, "GEN:%6li   MAX:%10.6f   MED:%10.6f   DEV:%10.6f   %02li:%02li:%02li  %02li:%02li:%02li\n",
          lGeracaoAtual, ga.statistics().current(GAStatistics::Maximum), ga.statistics().current(GAStatistics::Mean),
          ga.statistics().current(GAStatistics::Deviation), lTime / 3600, lTime % 3600 / 60, lTime % 60,
          lTimeRestante / 3600, lTimeRestante % 3600 / 60, lTimeRestante % 60);
      printf(szMensagem);
      if (bEcoTelaHabilitado) 
        fprintf(stderr, szMensagem);
    }
  }

  // Salva o melhor genoma no fim
  lTime = (long) (time(NULL) - tmInicial);
  printf("Tempo total de aprendizado: %02li:%02li:%02li\n", lTime / 3600, lTime % 3600 / 60, lTime % 60);
  ga.flushScores();
  genome = ga.statistics().bestIndividual();
  CarregarPesosFromGenoma(genome);
  SalvarPesos(sNomeArquivoBase + ".wts");
}


static void Initializer(GAGenome& g)
{
  GARealGenome & genome = (GARealGenome &) g;
  int i;

  // Inicializa os pesos na primeira geracao com valores pequenos
  for (i = 0; i < genome.size(); i++)
    genome.gene(i, GARandomFloat(-dInitPesos, dInitPesos));
}


static float Objective(GAGenome & g)
{
  double dTempoSimulado, dFatorBumpers, dDeslocamento, dInstabilidade, dFitness;
  static double dMelhorFitness = 0.0;

  // Executa a simulacao e calcula o fitness
  CarregarPesosFromGenoma(g);
  dTempoSimulado = PhysicalSimLoop(dTimeSimulation);
  dDeslocamento = CalcularDeslocamento();
  if (dDeslocamento > -5.0 * dTempoSimulado && dDeslocamento < 5.0 * dTempoSimulado) {
    dInstabilidade = CalcularInstabilidade();
    dFatorBumpers = CalcularFatorBumpers();
    dFitness = dDeslocamento / (1.0 + dFatorBumpers + PESO_GYRO * dInstabilidade);
  }
  else {
    dFitness = dDeslocamento = 0.0;
    dInstabilidade = dFatorBumpers = 1.0;
  }

  // Exibe os melhores fitness na tela e salva em um arquivo
  if (dFitness > dMelhorFitness) {
    dMelhorFitness = dFitness;
    lEpochOfBest = lGeracaoAtual;
    SalvarPesos(sNomeArquivoBase + ".wts");
    printf("FITN:%10.6f   DESL:%10.6f   BUMP:%9.6f   INST:%10.6f   TIME:%6.2f\n", dFitness, dDeslocamento,
        dFatorBumpers, dInstabilidade, dTempoSimulado);
  }
  // Seta a posicao original e retorna o fitness
  SetaPosOriginal();
  return (float) dFitness;
}


/*---------------------------------------------------------------------------------------------------------------------*
 * Funcoes Utilitarias                                                                                                 *
 *---------------------------------------------------------------------------------------------------------------------*/
string GetParamInFile(const string sNameFile, const string sParam)
{
  FILE *fp = NULL;
  char vcLinha[MAX_LINHA + 1];
  char *szPalavra;

  // Busca parametros no arquivo de configuracao
  if ((fp = fopen(sNameFile.c_str(), "r")) == NULL) {
    printf("Nao foi possivel abrir o arquivo de parametros\n");
    exit(1);
  }
  while (!feof(fp)) {
    if (fgets(vcLinha, MAX_LINHA, fp) && strlen(vcLinha) > 0) {
      if (vcLinha[0] == '#')
        continue;
      szPalavra = strtok(vcLinha, " ");
      while (szPalavra != NULL) {
        if (!strcmp(szPalavra, sParam.c_str())) {
          if ((szPalavra = strtok('\0', " "))) {
            fclose(fp);
            return szPalavra;
          }
          else
            continue;
        }
        szPalavra = strtok('\0', " ");
      }
    }
  }
  fclose(fp);
  printf("Parametro nao encontrado no arquivo de parametros\n");
  exit(1);
}


inline double DegToRad(const double dValDeg)
{
  const double MULTIP = M_PI / 180.0;
  return dValDeg * MULTIP;
}


inline double RadToDeg(const double dValRad)
{
  const double MULTIP = 180.0 / M_PI;
  return dValRad * MULTIP;
}


inline double Quadrado(const double dValor)
{
  return dValor * dValor;
}


inline double Normalizar(const double dValor, const double dMin, const double dMax)
{
  return (dValor - dMin) * 2.0 / (dMax - dMin) - 1.0;
}


inline double Anormaliza(const double dValor, const double dMin, const double dMax)
{
  return ((dValor + 1.0) * (dMax - dMin)) / 2.0 + dMin;
}


void HabilitaEcoTela(int iSinal) {
  if (iSinal == SIGUSR1) {
    fprintf(stderr, "Sinal SIGUSR1 recebido, eco em stderr habilitado\n");
    bEcoTelaHabilitado = true;
  }
  else if (iSinal == SIGUSR2) {
    fprintf(stderr, "Sinal SIGUSR2 recebido, eco em stderr desabilitado\n");
    bEcoTelaHabilitado = false;
  }
}


/*---------------------------------------------------------------------------------------------------------------------*
 * Funcoes de finalizacao                                                                                              *
 *---------------------------------------------------------------------------------------------------------------------*/
void EncerrarSimulacao()
{
  DestruirUniverso();
  LiberarMemoriaRobotnik();
  LiberarMemoriaRNA();
  if (bSimulacaoVisual)
    printf("\n");
  exit(0);
}


void TrataPedidoDestruicao(int iSinal) {
  static int iCount = 0;
  
  // Encerra o programa naturalmente quando o usuario der um kill
  if (iSinal == SIGTERM) {
    fprintf(stderr, "Sinal SIGTERM recebido, a aplicacao foi ABORTADA instantaneamente\n");
    exit(0);
  }
  else {
    iCount++;
    if (iCount > 1) {
      fprintf(stderr, "Sinal SIGINT recebido %d vezes, a aplicacao sera ABORTADA agora mesmo!\n", iCount);
      exit(0);
    }
    else {
      fprintf(stderr, "Sinal SIGINT recebido uma vez, AGUARDE o encerramento da aplicacao...\n");
      bAplicacaoCancelada = true;
    }
  }
}
